(* $Id: markup_aux.ml,v 1.10.2.1 2000/05/14 13:43:52 gerd Exp $
 * ----------------------------------------------------------------------
 * Markup! The validating XML parser for Objective Caml.
 * Copyright 1999 by Gerd Stolpmann. See LICENSE for details.
 * Some auxiliary functions 
 *)

(* TODO:
 * - Non Latin 1 characters:
 *   The function "character" should raise a special exception in this case;
 *   this should be trapped by any caller.
 *)


(**********************************************************************)
(* Lexing *)


open Markup_types
open Markup_lexer_types

let character warner k =
  if k < 0 then raise (Illegal_character 0);
  if k = 9 or k = 10 or k = 13 or (k >= 32 & k <= 255) then
    String.make 1 (Char.chr k)
  else 
    if k>=256 then
      begin
	if k < 0xd800 or (k >= 0xe000 & k <= 0xfffd) or 
	   (k >= 0x10000 & k <= 0x10ffff) then begin
	     warner # warn ("Character not in Latin 1 alphabet: " ^ string_of_int k);
	     ""
	   end
	else
	  raise(Illegal_character 0)
      end
    else
      raise(Illegal_character 0)
;;


let check_name warner name =
  (* produces a warning for names beginning with "xml". *)
  if String.length name >= 3 then begin
    match String.sub name 0 3 with
	("xml" | "xmL" | "xMl" | "xML" | "Xml" | "XmL" | "XMl" | "XML") ->
	  warner # warn ("Name is reserved for future extensions: " ^ name)
      | _ ->
	  ()
  end
;;


let tokens_of_content_string s =
  (* tokenizes general entities and character entities *)
  let lexbuf = Lexing.from_string s in
  let rec next_token () =
    match Markup_lexers.scan_content_string lexbuf with
	Eof -> []
      | tok -> tok :: next_token()
  in
  next_token()
;;


let rec expand_attvalue_with_rec_check dtd s warner entities norm_crlf =
  (* recursively expands general entities and character entities;
   * normalizes whitespace
   *)
  let toklist = tokens_of_content_string s in
  let rec expand tl =
    match tl with
	[] -> ""
      | ERef n :: tl' ->
	  if List.mem n entities then
	    raise(WF_error("Recursive reference to general entity `" ^ n ^ "'"));
	  let en = dtd # gen_entity n in
	  let rtext, rtext_contains_ext_refs = en # replacement_text in
	  if rtext_contains_ext_refs then
	    raise(Validation_error("Found reference to external entity in attribute value"));
	  expand_attvalue_with_rec_check 
	    dtd rtext warner (n :: entities) false    ^    expand tl'
      | CRef(-1) :: tl' ->
	  if norm_crlf then
	    " " ^ expand tl'
	  else
	    "  " ^ expand tl'
      | CRef n :: tl' ->
	  character warner n ^ expand tl'
      | CharData "<" :: tl' ->
	  raise 
	    (WF_error
	       ("Attribute value contains character '<' literally"))
      | CharData x :: tl' ->
	  x ^ expand tl'
      | _ -> assert false
  in
  expand toklist
;;


let expand_attvalue dtd s warner norm_crlf =
  (* norm_crlf: whether the sequence CRLF is recognized as one character or
   * not (i.e. two characters)
   *)
  expand_attvalue_with_rec_check dtd s warner [] norm_crlf
;;


let count_lines s =
  (* returns number of lines in s, number of columns of the last line *)
  let l = String.length s in

  let rec count n k no_cr no_lf =
    let next_cr = 
      if no_cr then
	(-1)
      else
	try String.index_from s k '\013' with Not_found -> (-1) in
    let next_lf = 
      if no_lf then
	(-1)
      else
	try String.index_from s k '\010' with Not_found -> (-1) in
    if next_cr >= 0 & (next_lf < 0 or next_cr < next_lf) then begin
      if next_cr+1 < l & s.[next_cr+1] = '\010' then
	count (n+1) (next_cr+2) false (next_lf < 0)
      else
	count (n+1) (next_cr+1) false (next_lf < 0)
    end
    else if next_lf >= 0 then begin
      count (n+1) (next_lf+1) (next_cr < 0) false
    end
    else
      n, (l - k)

  in
  count 0 0 false false
;;


let decode_xml_pi pl =
  (* 'pl' must consist of name="value" or name='value' pairs which are returned
   * as list of pairs.
   * The "value" is returned as it is; no substitution of &entities; happens.
   *)
  let rec decode pl =
    match pl with
	Pro_name name :: Pro_eq :: Pro_string value :: pl' ->
	  (name, value) :: decode pl'
      | [] ->
	  []
      | _ ->
	  raise (WF_error("Bad XML processing instruction"))
  in
  decode pl
;;


let decode_doc_xml_pi pl =
  match pl with
      [ "version", v ]                                  -> (v, None, None)
    | [ "version", v; "encoding", e ]                   -> (v, Some e, None)
    | [ "version", v; "standalone", s ]                 -> (v, None, Some s)
    | [ "version", v; "encoding", e; "standalone", s ]  -> (v, Some e, Some s)
    | _ ->
	raise(WF_error("Bad XML declaration"))
;;


let check_text_xml_pi pl =
  match pl with
    | [ "version", v; "encoding", e ] -> ()
    | [ "encoding", e ]  -> ()
    | _ ->
	raise(WF_error("Bad XML declaration"))
;;


let check_version_num s =
  let l = String.length s in
  for i = 0 to l - 1 do
    match s.[i] with
	('a'..'z'|'A'..'Z'|'0'..'9'|
	 '-'|'_'|'.'|':') -> ()
      | _ ->
	  raise(WF_error("Bad XML version string"))
  done
;;


let check_public_id s =
  let l = String.length s in
  for i = 0 to l - 1 do
    match s.[i] with
	(' '|'\013'|'\010'|'a'..'z'|'A'..'Z'|'0'..'9'|
	 '-'|'\''|'('|')'|'+'|','|'.'|'/'|':'|'='|'?'|
	 ';'|'!'|'*'|'#'|'@'|'$'|'_'|'%') -> ()
      | _ ->
	  raise(WF_error("Illegal character in PUBLIC identifier"))
  done
;;


(**********************************************************************)
(* list functions *)


let rec check_dups l =
  match l with
      [] -> false
    | c :: l' -> 
	if List.mem c l' then true else check_dups l'
;;


let rec count pred l =
  match l with
      [] -> 0
    | x :: l' -> 
	if pred x then  1 + (count pred l') else count pred l'
;;


(**********************************************************************)
(* attributes *)

let check_attribute_value_lexically x t v =
  (* raises x if the attribute value v does not match the lexical rules
   * for attribute type t:
   * - t = A_id: v must be a <name>
   * - t = A_idref: v must match <name>
   * - t = A_idrefs: v must match <names>
   * - t = A_entity: v must match <name>
   * - t = A_entities: v must match <names>
   * - t = A_nmtoken: v must match <nmtoken>
   * - t = A_nmtokens: v must match <nmtokens>
   * - other types not checked
   *)
  let lexbuf = Lexing.from_string v in
  let rec get_name_list() =
    match Markup_lexers.scan_name_string lexbuf with
	Eof -> []
      | tok -> tok :: get_name_list()
  in
  let l = get_name_list() in
  match t with
      (A_id | A_idref | A_entity) ->
	begin match l with
	    [ Name n ] -> ()
	  | _          -> raise (Lazy.force x)
	end
    | (A_idrefs | A_entities) ->
	if List.exists (fun tok -> 
			  match tok with
			      Name _ -> false
			    | Ignore -> false
			    | _ -> true) l then
	  raise (Lazy.force x)
    | A_nmtoken ->
	begin match l with
	    [ Name n ]      -> ()
	  | [ Nametoken n ] -> ()
	  | _               -> raise (Lazy.force x)
	end
    | A_nmtokens ->
	if List.exists (fun tok -> 
			  match tok with
			      Name _ -> false
			    | Nametoken _ -> false
			    | Ignore -> false
			    | _ -> true
		       ) l then
	  raise (Lazy.force x)
    | _ -> ()
;;


let split_attribute_value v =
  (* splits 'v' into a list of names or nmtokens *)
  let lexbuf = Lexing.from_string v in
  let rec get_name_list() =
    match Markup_lexers.scan_name_string lexbuf with
	Eof         -> []
      | Ignore      -> get_name_list()
      | Name s      -> s :: get_name_list()
      | Nametoken s -> s :: get_name_list()
      | _           -> raise(Validation_error("Illegal attribute value"))
  in
  get_name_list()
;;


let normalize_line_separators s =
  let lexbuf = Lexing.from_string s in
  let rec get_string() =
    match Markup_lexers.scan_for_crlf lexbuf with
	Eof        -> ""
      | CharData s -> s ^ get_string()
      | _          -> assert false
  in
  get_string()
;;


(**********************************************************************)

let write_data_string os content =
  (* TODO: It's a bit slow. *)
  let i = ref 0 in
  for k = 0 to String.length content - 1 do
    match content.[k] with
	('&' | '<' | '>' | '"' | '%') as c ->
	  if !i < k then
	    write os content !i (k - !i);
	  begin match c with
	      '&' -> write os "&amp;"  0 5
	    | '<' -> write os "&lt;"   0 4
	    | '>' -> write os "&gt;"   0 4
	    | '"' -> write os "&quot;" 0 6
	    | '%' -> write os "&#37;"  0 5     (* reserved in DTDs *)
	    | _   -> assert false
	  end;
	  i := k+1
      | _ -> ()
  done;
  if !i < String.length content then
    write os content !i (String.length content - !i)
;;


(* ======================================================================
 * History:
 * 
 * $Log: markup_aux.ml,v $
 * Revision 1.10.2.1  2000/05/14 13:43:52  gerd
 * 	write_data_string: The character '%' is special, too.
 *
 * Revision 1.10  2000/05/01 20:41:56  gerd
 * 	New function write_data_string.
 *
 * Revision 1.9  2000/04/30 18:11:31  gerd
 * 	New function normalize_line_separators.
 * 	In function expand_attvalue: New argument norm_crlf. If the attvalue
 * is read directly from a file, the sequence CR LF must be converted to a
 * single space. If the attvalue is read from a replacement text, CR LF has
 * already converted to a single LF, and CR LF, if still occurring, must be
 * converted to two spaces. The caller can indicate the case by passing
 * true/false as norm_crlf.
 *
 * Revision 1.8  1999/09/01 22:51:07  gerd
 * 	Added functions.
 * 	'character' raises Illegal_character if characters are found that
 * do not match the production Char.
 *
 * Revision 1.7  1999/09/01 16:17:37  gerd
 * 	Added function 'check_name'.
 *
 * Revision 1.6  1999/08/15 20:33:19  gerd
 * 	Added: a function that checks public identifiers. Only certain
 * characters may occur in these identifiers.
 * 	Control characters are rejected by the "character" function.
 * 	Bugfix: recursive entity references are detected in attribute
 * expansion
 *
 * Revision 1.5  1999/08/15 02:18:02  gerd
 * 	That '<' is not allowed in attribute values, is a violation
 * of well-formedness, not of the validity; so WF_error is raised.
 *
 * Revision 1.4  1999/08/15 00:20:37  gerd
 * 	When expanding attribute values, references to parameter
 * entities are now resolved by the method "replacement_text" which
 * has an additional return value, and no longer by "attlist_replacement_text".
 * The new return value indicates whether references to external entities
 * have been resolved (directly or indirectly); this is allowed at some
 * locations but not in attribute values.
 *
 * Revision 1.3  1999/08/14 22:05:53  gerd
 * 	Several functions have now a "warner" as argument which is
 * an object with a "warn" method. This is used to warn about characters
 * that cannot be represented in the Latin 1 alphabet.
 *
 * Revision 1.2  1999/08/10 21:35:06  gerd
 * 	The XML/encoding declaration at the beginning of entities is
 * evaluated. In particular, entities have now a method "xml_declaration"
 * which returns the name/value pairs of such a declaration. The "encoding"
 * setting is interpreted by the entity itself; "version", and "standalone"
 * are interpreted by Markup_yacc.parse_document_entity. Other settings
 * are ignored (this does not conform to the standard; the standard prescribes
 * that "version" MUST be given in the declaration of document; "standalone"
 * and "encoding" CAN be declared; no other settings are allowed).
 * 	TODO: The user should be warned if the standard is not exactly
 * fulfilled. -- The "standalone" property is not checked yet.
 *
 * Revision 1.1  1999/08/10 00:35:50  gerd
 * 	Initial revision.
 *
 * 
 *)
