(* $Id: markup_entity.ml,v 1.19 2000/05/01 15:18:44 gerd Exp $
 * ----------------------------------------------------------------------
 * Markup! The validating XML parser for Objective Caml.
 * Copyright 1999 by Gerd Stolpmann. See LICENSE for details.
 *)


open Markup_types
open Markup_lexer_types
open Markup_aux
open Markup_reader

(* Hierarchy of parsing layers:
 *
 * - Parser: Markup_yacc
 *   + gets input stream from the main entity object
 *   + checks most of the grammar
 *   + creates the DTD object as side-effect
 *   + creates the element tree as side-effect
 *   + creates further entity objects that are entered into the DTD
 * - Entity layer: Markup_entity
 *   + gets input stream from the lexers, or another entity object
 *   + handles entity references: if a reference is encountered the
 *     input stream is redirected such that the tokens come from the
 *     referenced entity object
 *   + handles conditional sections
 * - Lexer layer: Markup_lexers
 *   + gets input from lexbuffers created by resolvers
 *   + different lexers for different lexical contexts
 *   + a lexer returns pairs (token,lexid), where token is the scanned
 *     token, and lexid is the name of the lexer that must be used for
 *     the next token
 * - Resolver layer: Markup_entity
 *   + a resolver creates the lexbuf from some character source
 *   + a resolver recodes the input and handles the encoding scheme
 *)

(**********************************************************************)

(* Variables of type 'state' are used to insert Begin_entity and End_entity
 * tokens into the stream.
 * - At_beginning: Nothing has been read so far
 * - First_token tok: A Begin_entity has been inserted; and the next token
 *   is 'tok' which is not Eof. (Begin_entity/End_entity must not be inserted
 *   if the entity is empty.)
 * - In_stream: After the first token has been read, but befor Eof.
 * - At_end: Eof has been read, and End_entity has been returned.
 *)

type state =
    At_beginning
  | First_token of token
  | In_stream
  | At_end
;;

(**********************************************************************)

class virtual entity the_dtd the_name the_warner init_errors_with_line_numbers =
  object (self)
    (* This class prescribes the type of all entity objects. Furthermore,
     * the default 'next_token' mechanism is implemented.
     *)

    (* 'init_errors_with_line_numbers': whether error messages contain line
     * numbers or not.
     * Calculating line numbers is expensive.
     *)

    val mutable dtd = the_dtd
    val mutable name = the_name
    val mutable warner = the_warner

    val mutable lexbuf = Lexing.from_string ""
      (* The lexical buffer currently used as character source. *)

    val mutable prolog = None
      (* Stores the initial <?xml ...?> token as PI_xml *)

    val mutable prolog_pairs = []
      (* If prolog <> None, these are the (name,value) pairs of the
       * processing instruction.
       *)

    val mutable accept_prolog = true
      (* 'true' as long as <?xml ...?> is accepted *)

    val mutable lex_id = Document
      (* The name of the lexer that should be used for the next token *)

    val mutable delegated_entity = None
      (* None:    The entity reads currently tokens from lexbuf
       * Some e:  The entity reads currently tokens from the other entity e.
       *          This mode is used to resolve entity references. It is
       *          enabled once a &name; or %name; has been seen, and lasts
       *          until e returns Eof. Note that this final Eof is suppressed
       *          and not passed over to the caller.
       * TODO: This way of resolving entity references is simple but slows
       * the parser down if deeply nested references must be processed,
       * because every token must be passed from the innermost entity to the
       * outermost entity. It is much better to install a central entity stack
       * for this purpose. The outermost entity should have this stack. A
       * special token Push(new_entity) is used if an entity wants to redirect
       * the input to new_entity; Eof can be used to recognize the end of a
       * redirection.
       *)

    val mutable cond_level = 0
      (* How many beginnings of (included) conditional sections have been seen.
       *)

    val mutable allow_conditionals = true
      (* true iff conditional sections are allowed in this entity *)

    val mutable force_parameter_entity_parsing = false
      (* 'true' forces that inner entities will always be embraced by
       *        Begin_entity and End_entity.
       * 'false': the inner entity itself decides this
       *)

    val mutable check_text_declaration = true
      (* 'true': It is checked that the <?xml..?> declaration matches the
       *         production TextDecl.
       *)

    val mutable normalize_newline = true
      (* Whether this entity converts CRLF or CR to LF, or not *)


    val mutable line = 1     (* current line *)
    val mutable column = 0   (* current column *)
    val mutable pos = 0      (* current absolute character position *)
    val errors_with_line_numbers = init_errors_with_line_numbers

    val mutable last_token = Eof
    val mutable space_seen = false
      (* These two variables are used to check that between certain pairs of
       * tokens whitespaces exist. 'last_token' is simply the last token,
       * but not Ignore, and not PERef (which both represent whitespace).
       * 'space_seen' records whether Ignore or PERef was seen between this
       * token and 'last_token'.
       *)

    val mutable debug = false

    method is_ndata = false
      (* Returns if this entity is an NDATA (unparsed) entity *)

    method name = name

    method virtual open_entity : bool -> lexers -> unit
	(* open_entity force_parsing lexid:
	 * opens the entity, and the first token is scanned by the lexer
	 * 'lexid'. 'force_parsing' forces that Begin_entity and End_entity
	 * tokens embrace the inner tokens of the entity; otherwise this
	 * depends on the entity.
	 * By opening an entity, reading tokens from it, and finally closing
	 * the entity, the inclusion methods "Included",
	 * "Included if validating", and "Included as PE" can be carried out.
	 * Which method is chosen depends on the 'lexid', i.e. the lexical
	 * context: 'lexid = Content' performs "Included (if validating)" (we
	 * are always validating); 'lexid = Declaration' performs
	 * "Included as PE". The difference is which tokens are recognized,
	 * and how spaces are handled.
	 * 'force_parsing' causes that a Begin_entity token is inserted before
	 * and an End_entity token is inserted after the entity. The yacc
	 * rules allow the Begin_entity ... End_entity brace only at certain
	 * positions; this is used to restrict the possible positions where
	 * entities may be included, and to guarantee that the entity matches
	 * a certain production of the grammar ("parsed entities").
	 * 'open_entity' is currently invoked with 'force_parsing = true'
	 * for toplevel nodes, for inclusion of internal general entities,
	 * and for inclusion of parameter entities into document entities.
	 * 'force_parsing = false' is used for all other cases: External
	 * entities add the Begin_entity/End_entity tokens anyway; internal
	 * entities do not. Especially internal parameter entities referenced
	 * from non-document entities do not add these tokens.
	 *)

    method virtual close_entity : lexers
	(* close_entity:
	 * closes the entity and returns the name of the lexer that must
	 * be used to scan the next token.
	 *)

    method virtual replacement_text : (string * bool)
	(* replacement_text:
	 * returns the replacement text of the entity, and as second value,
	 * whether the replacement text was constructed by referencing
	 * external entities (directly or indirectly).
	 * This method implements the inclusion method "Included in Literal".
	 *)


    method lexbuf = lexbuf


    method xml_declaration =
      (* return the (name,value) pairs of the initial <?xml name=value ...?>
       * processing instruction.
       *)
      match prolog with
	  None ->
	    None
	| Some p ->
	    Some prolog_pairs


    method set_debugging_mode m =
      debug <- m

    method private virtual set_encoding : string -> unit


    method position_string =
      (* Gets a string describing the position of the last token *)
      let n = self # full_name in
      match delegated_entity with
	  None ->
	    "In entity " ^ n ^ ", at line " ^ string_of_int line ^
	    ", column " ^ string_of_int column ^ ": "
	| Some e ->
	    "Called from entity " ^ n ^ ", line " ^ string_of_int line ^
	    ", column " ^ string_of_int column ^ ":\n" ^
	    e # position_string


    method private full_name =
      name


    method next_token =
      (* Get the next token of the token stream. Entity references and
       * conditional sections are already processed by this method; this
       * is transparent for the caller (except that Begin_entity and End_entity
       * tokens are sometimes added to the stream).
       *)
      let rec read_token() =
	(* read next token and interpret conditional sections *)
	(* Important invariant:
         * Conditional_begin, and Conditional_end are never
         * returned to the caller of 'next_token'; every entity processes
         * conditional sections itself (i.e. conditional sections are not
         * allowed to begin in one entity and end in another entity).
         *)
	let tok = self # next_token_from_entity() in
	match tok with
	    Conditional_begin ->
	      if not allow_conditionals then
		raise(WF_error("Restriction of the internal subset: Conditional and CDATA sections not allowed"));
	      let tok' = self # next_token_from_entity() in
	      begin match tok' with
		  Name "INCLUDE" ->
		    cond_level <- cond_level + 1;
		    let tok'' = self # next_token_from_entity() in
		    if tok'' <> Conditional_body then
		      raise Parsing.Parse_error;
		    read_token()
		| Name "IGNORE" ->
		    let tok'' = self # next_token_from_entity() in
		    if tok'' <> Conditional_body then
		      raise Parsing.Parse_error;
		    let llev = ref 1 in
		    while !llev >= 1 do
		      let igntok = self # next_ignored_token() in
		      (* next_ignored_token: uses a special lexer that only
		       * recognizes Conditional_begin and Conditional_end;
		       * other character combinations are ignored.
		       *)
		      match igntok with
			  Conditional_begin ->
			    llev := !llev + 1
			| Conditional_end ->
			    llev := !llev - 1
			| Eof ->
			    raise Parsing.Parse_error
			| _ ->
			    ()
		    done;
		    read_token()
		| _ -> raise Parsing.Parse_error
	      end

	  | Conditional_end ->
	      if cond_level = 0 then raise Parsing.Parse_error;
	      cond_level <- cond_level - 1;
	      read_token()

	  | Conditional_body ->
	      Conditional_body

	  | Eof ->
	      if cond_level <> 0 then raise Parsing.Parse_error;
	      Eof

	  | _ -> tok
		(* The default: pass the token back to the caller *)
      in
      read_token()


    method private next_token_from_entity() =
      (* read next token from this or from the delegated entity *)

      match delegated_entity with
	  None -> self # next_token_from_this_entity()
                  (* The token stream comes from *this* entity *)
	| Some e ->
            (* e is the other entity that is temporarily the source of tokens *)
	    let tok = e # next_token in
	    match tok with
		Eof -> lex_id <- e # close_entity;
		       delegated_entity <- None;
		       self # next_token_from_this_entity()
	      | _   -> tok

    method private next_token_from_this_entity () =
      (* read next token from this entity *)

        let this_line = line
        and this_column = column in
	let this_pos = pos in
	(* Read the next token from the appropriate lexer lex_id, and get the
	 * name lex_id' of the next lexer to be used.
	 *)
	let tok, lex_id' =
	  match lex_id with
	      Document      -> Markup_lexers.scan_document lexbuf
	    | Document_type -> Markup_lexers.scan_document_type lexbuf
	    | Content       -> Markup_lexers.scan_content lexbuf
	    | Within_tag    -> Markup_lexers.scan_within_tag lexbuf
	    | Declaration   -> Markup_lexers.scan_declaration lexbuf
	    | Ignored_section -> assert false
		  (* Ignored_section: only used by method next_ignored_token *)
	in
	if debug then
	  prerr_endline ("- Entity " ^ name ^ ": " ^ string_of_tok tok);
	(* Find out the number of lines and characters of the last line: *)
	let n_lines, n_columns =
	  if errors_with_line_numbers then
	    count_lines (Lexing.lexeme lexbuf)
	  else
	    0, (Lexing.lexeme_end lexbuf - Lexing.lexeme_start lexbuf)
	in
	line <- this_line + n_lines;
	column <- if n_lines = 0 then this_column + n_columns else n_columns;
	pos <- Lexing.lexeme_end lexbuf;
	lex_id <- lex_id';
	(* Throw Ignore and Comment away; Interpret entity references: *)
	(* NOTE: Of course, references to general entities are not allowed
	 * everywhere; parameter references, too. This is already done by the
	 * lexers, i.e. &name; and %name; are recognized only where they
	 * are allowed.
	 *)
	let old_accept_prolog = accept_prolog in
	accept_prolog <- false;
	(* Note: 'accept_prolog' is true only if this is the first token of
	 * the entity!
	 *)
	match tok with
	    Ignore    -> if old_accept_prolog then
	                   self # set_encoding "";
	                 space_seen <- true;
                         self # next_token_from_this_entity()
	  | Comment   -> if old_accept_prolog then
	                   self # set_encoding "";
	                 space_seen <- false;
			 last_token <- Comment;
                         self # next_token_from_this_entity()
	  | PI_xml pl -> if old_accept_prolog then begin
	                   prolog <- Some pl;
	                   prolog_pairs <- decode_xml_pi pl;
			   if check_text_declaration then
			     check_text_xml_pi prolog_pairs;
			   begin
			     try
			       let e = List.assoc "encoding" prolog_pairs in
			       self # set_encoding e
			     with
				 Not_found ->
				   self # set_encoding ""
			   end;
			   space_seen <- false;
			   last_token <- tok;
			   self # next_token_from_this_entity()
	                 end
	                 else raise(WF_error("Prolog not allowed here"))
	  | ERef n    -> if old_accept_prolog then
	                   self # set_encoding "";
	                 space_seen <- false;
			 last_token <- tok;
                         let en = dtd # gen_entity n in
			 en # set_debugging_mode debug;
	                 en # open_entity true lex_id;
	                 delegated_entity <- Some en;
		         self # next_token_from_entity()
	  | PERef n   -> if old_accept_prolog then
	                   self # set_encoding "";
	                 space_seen <- true;
                         let en = dtd # par_entity n in
			 en # set_debugging_mode debug;
	                 en # open_entity force_parameter_entity_parsing lex_id;
	                 delegated_entity <- Some en;
		         self # next_token_from_entity()

          (* Convert LineEnd to CharData *)
	  | LineEnd s -> if old_accept_prolog then
	                   self # set_encoding "";
	                 space_seen <- false;
			 let new_tok =
			   if normalize_newline then 
			     CharData "\n"
			   else
			     CharData s
			 in
			 last_token <- new_tok;
                         new_tok

          (* Also normalize CDATA sections *)
	  | Cdata value as cd ->
	                 if old_accept_prolog then
	                   self # set_encoding "";
	                 space_seen <- false;
			 let new_tok =
			   if normalize_newline then 
			     Cdata(normalize_line_separators value)
			   else
			     cd
			 in
			 last_token <- new_tok;
                         new_tok

          (* If there are CRLF sequences in a PI value, normalize them, too *)
	  | PI(name,value) as pi ->
	                 if old_accept_prolog then
	                   self # set_encoding "";
	                 space_seen <- false;
			 let new_tok =
			   if normalize_newline then
			     PI(name, normalize_line_separators value)
			   else
			     pi
			 in
			 last_token <- new_tok;
                         new_tok
         
          (* Attribute values: If they are already normalized, they are turned
	   * into Attval_nl_normalized. This is detected by other code.
	   *)
	  | Attval value as av ->
	                 if old_accept_prolog then
			   self # set_encoding "";
	                 space_seen <- false;
			 let new_tok =
			   if normalize_newline then
			     av
			   else
			     Attval_nl_normalized value
			 in
			 last_token <- new_tok;
                         new_tok

          (* Another CRLF normalization case: See Unparsed_string below. *)


	  (* The following cases are only handled separately because of
	   * white space checking.
	   *)

	  | Name _ ->    if old_accept_prolog then
	                   self # set_encoding "";
	                 if not space_seen &
			    match last_token with
				(Decl_element _|Decl_attlist _|Decl_entity _|
				 Decl_notation _|Doctype _|Percent) -> true
			      | Name _ -> true
			      | Nametoken _ -> true
			      | Attval _ -> true
			      | Unparsed_string _ -> true
			      | _ -> false
			 then
			   raise(WF_error("Here must be whitespace"));
			 space_seen <- false;
			 last_token <- tok;
			 tok
	  | Percent ->   if old_accept_prolog then
	                   self # set_encoding "";
	                 if not space_seen &
			    match last_token with
				Decl_entity _ -> true
			      | _ -> false
			 then
			   raise(WF_error("Here must be whitespace"));
			 space_seen <- false;
			 last_token <- tok;
			 tok
	  | Lparen _ ->  if old_accept_prolog then
	                   self # set_encoding "";
	                 if not space_seen &
			    match last_token with
				Name _ -> true
			      | _ -> false
			 then
			   raise(WF_error("Here must be whitespace"));
			 space_seen <- false;
			 last_token <- tok;
			 (* set entity_id: *)
			 Lparen (self :> entity_id)
	  | (Required|Implied|Fixed) ->
	                 if old_accept_prolog then
	                   self # set_encoding "";
	                 if not space_seen &
			    match last_token with
				Name _   -> true
			      | Rparen _ -> true
			      | _        -> false
			 then
			   raise(WF_error("Here must be whitespace"));
			 space_seen <- false;
			 last_token <- tok;
			 tok
	  | Unparsed_string value as ustr ->
	                 if old_accept_prolog then
	                   self # set_encoding "";
	                 if not space_seen &
			    match last_token with
				Name _   -> true
			      | Rparen _ -> true
			      | Fixed    -> true
			      | Unparsed_string _ -> true
			      | _        -> false
			 then
			   raise(WF_error("Here must be whitespace"));
			 space_seen <- false;
			 let new_tok =
			   if normalize_newline then
			     Unparsed_string(normalize_line_separators value)
			   else
			     ustr
			 in
			 last_token <- new_tok;
			 new_tok
	  | (Plus|Star|Qmark) ->
	                 if old_accept_prolog then
	                   self # set_encoding "";
	                 if space_seen then
			   raise(WF_error("Whitespace not allowed here"));
			 space_seen <- false;
			 last_token <- tok;
			 tok

          (* These token require that the entity_id parameter is set: *)
          | (Doctype _|Doctype_rangle _|Dtd_begin _|Dtd_end _|Decl_element _|
	     Decl_attlist _|Decl_entity _|Decl_notation _|Decl_rangle _|
	     Rparen _|RparenPlus _|RparenStar _|RparenQmark _|
	     Tag_beg(_,_)|Tag_end(_,_)) ->
	                 if old_accept_prolog then
	                   self # set_encoding "";
	                 space_seen <- false;
			 last_token <- tok;
			 begin match tok with
			     Doctype _      -> Doctype       (self :> entity_id)
			   |Doctype_rangle _ ->Doctype_rangle(self :> entity_id)
			   | Dtd_begin _    -> Dtd_begin     (self :> entity_id)
			   | Dtd_end _      -> Dtd_end       (self :> entity_id)
			   | Decl_element _ -> Decl_element  (self :> entity_id)
			   | Decl_attlist _ -> Decl_attlist  (self :> entity_id)
			   | Decl_entity _  -> Decl_entity   (self :> entity_id)
			   | Decl_notation _ ->Decl_notation (self :> entity_id)
			   | Decl_rangle _  -> Decl_rangle   (self :> entity_id)
			   | Rparen _       -> Rparen        (self :> entity_id)
			   | RparenPlus _   -> RparenPlus    (self :> entity_id)
			   | RparenStar _   -> RparenStar    (self :> entity_id)
			   | RparenQmark _  -> RparenQmark   (self :> entity_id)
			   | Tag_beg (n,_)  -> Tag_beg (n, (self :> entity_id))
			   | Tag_end (n,_)  -> Tag_end (n, (self :> entity_id))
			   | _ -> assert false
			 end

          (* The default case. *)

	  | _         -> if old_accept_prolog then
	                   self # set_encoding "";
	                 space_seen <- false;
			 last_token <- tok;
                         tok

    method private next_ignored_token() =
        (* used after <![ IGNORE *)
        let this_line = line
        and this_column = column in
	let this_pos = pos in
	let tok, lex_id' = Markup_lexers.scan_ignored_section lexbuf in
	let n_lines, n_columns = count_lines (Lexing.lexeme lexbuf) in
	line <- this_line + n_lines;
	column <- if n_lines = 0 then this_column + n_columns else n_columns;
	pos <- Lexing.lexeme_end lexbuf;
	tok

  end
;;


class ndata_entity the_name the_ext_id the_notation =
  object (self)
    (* An NDATA entity is very restricted; more or less you can only find out
     * its external ID and its notation.
     *)

    val mutable name = the_name
    val mutable ext_id = the_ext_id
    val mutable notation = the_notation

    method name = (name : string)
    method ext_id = (ext_id : ext_id)
    method notation = (notation : string)

    method is_ndata = true

    method private set_encoding (_:string) =
      assert false

    method xml_declaration = (None : (string*string) list option)

    method set_debugging_mode (_:bool) = ()

    method open_entity (_:bool) (_:lexers) =
      ( raise (Validation_error ("Invalid reference to NDATA entity " ^ name))
	  : unit )

    method close_entity =
      ( raise (Validation_error ("Invalid reference to NDATA entity " ^ name))
	  : lexers )

    method replacement_text =
      ( raise (Validation_error ("Invalid reference to NDATA entity " ^ name))
	  : (string * bool) )

    method lexbuf =
      ( raise (Validation_error ("Invalid reference to NDATA entity " ^ name))
	  : Lexing.lexbuf )

    method position_string =
      ( raise (Validation_error ("Invalid reference to NDATA entity " ^ name))
	  : string )

    method next_token =
      ( raise (Validation_error ("Invalid reference to NDATA entity " ^ name))
	  : token )
  end
;;


class external_entity the_resolver the_dtd the_name the_warner the_ext_id
                      the_p_special_empty_entities
		      init_errors_with_line_numbers
  =
  object (self)
    inherit entity
              the_dtd the_name the_warner init_errors_with_line_numbers
            as super

    (* An external entity gets the lexbuf that is used as character source
     * from a resolver.
     * Furthermore, before the first token an Begin_entity is inserted, and
     * before Eof an End_entity token is inserted into the stream. This done
     * always regardless of the argument 'force_parsing' of the method
     * 'open_entity'.
     *
     * 'the_p_internal_subset': see class internal_entity
     * 'the_p_special_empty_entities': if true, a Begin_entity/End_entity
     * brace is left out if the entity is otherwise empty.
     *)

    val resolver = (the_resolver : resolver)
    val ext_id = (the_ext_id : ext_id)

    val p_special_empty_entities = (the_p_special_empty_entities : bool)

    val mutable resolver_is_open = false
      (* Track if the resolver is open. This is also used to find recursive
       * references of entities.
       *)

    val mutable state = At_beginning


    method private set_encoding e =
      assert resolver_is_open;
      resolver # change_encoding e


    method private full_name =
      name ^
      match ext_id with
	  System s    -> " = SYSTEM \"" ^ s ^ "\""
	| Public(p,s) -> " = PUBLIC \"" ^ p ^ "\" \"" ^ s ^ "\""


    method open_entity force_parsing init_lex_id =
      (* Note that external entities are always parsed, i.e. Begin_entity
       * and End_entity tokens embrace the inner tokens to force that
       * the entity is only called where the syntax allows it.
       *)
      if resolver_is_open then
	raise(Validation_error("Recursive reference to entity `" ^ name ^ "'"));
      let lex = resolver # open_in ext_id in
      resolver_is_open <- true;
      lexbuf  <- lex;
      prolog  <- None;
      accept_prolog <- true;
      lex_id  <- init_lex_id;
      delegated_entity <- None;
      cond_level <- 0;
      state <- At_beginning;
      line <- 1;
      column <- 0;
      pos <- 0;
      space_seen <- false;
      last_token <- Eof;
      normalize_newline <- true;

    method next_token =
      if not resolver_is_open then
	failwith ("External entity " ^ name ^ " not open");
      let t =
	begin match state with
	    At_beginning ->
	      let tok = super # next_token in
	      if tok = Eof & p_special_empty_entities then begin
		state <- At_end;
		Eof
	      end else begin
		state <- First_token tok;
		Begin_entity
	      end
	  | First_token tok ->
	      ( match tok with
		    Eof -> state <- At_end; End_entity
		  | _   -> state <- In_stream; tok
	      )
	  | In_stream ->
	      let tok = super # next_token in
	      ( match tok with
		    Eof -> state <- At_end; End_entity
		  | _   -> tok
	      )
	  | At_end -> Eof
	end
      in
      t

    method close_entity =
      if not resolver_is_open then
	failwith ("External entity " ^ name ^ " not open");
      resolver # close_in;
      resolver_is_open <- false;
      lex_id

    method replacement_text =
      (* Return the replacement text of the entity. The method used for this
       * is more or less the same as for internal entities; i.e. character
       * and parameter entities are resolved immediately. In addition to that,
       * external entities may begin with an "xml" processing instruction
       * which is considered not to be part of the replacement text.
       *)
      if resolver_is_open then
	raise(Validation_error("Recursive reference to entity `" ^ name ^ "'"));
      let lex = resolver # open_in ext_id in
      resolver_is_open <- true;
      lexbuf  <- lex;
      prolog  <- None;
      accept_prolog <- true;
      (* arbitrary:    lex_id  <- init_lex_id; *)
      delegated_entity <- None;
      cond_level <- 0;
      state <- At_beginning;
      line <- 1;
      column <- 0;
      pos <- 0;
      space_seen <- false;
      last_token <- Eof;
      (* First check if the first token of 'lex' is <?xml...?> *)
      begin match Markup_lexers.scan_only_xml_decl lex with
	  PI_xml pl ->
	    prolog <- Some pl;
	    prolog_pairs <- decode_xml_pi pl;
	    begin
	      try
		let e = List.assoc "encoding" prolog_pairs in
		self # set_encoding e
	      with
		  Not_found ->
		    self # set_encoding ""
	    end;
	| Eof ->
	    (* This only means that the first token was not <?xml...?>;
	     * the "Eof" token represents the empty string.
	     *)
	    ()
	| _ ->
	    (* Must not happen. *)
	    assert false
      end;
      (* Then create the replacement text. *)
      let rec scan_and_expand () =
	match Markup_lexers.scan_dtd_string lexbuf with
	    ERef n -> "&" ^ n ^ ";" ^ scan_and_expand()
	  | CRef(-1) -> "\n" ^ scan_and_expand()
	  | CRef(-2) -> "\n" ^ scan_and_expand()
	  | CRef(-3) -> "\n" ^ scan_and_expand()
	  | CRef k -> character warner k ^ scan_and_expand()
	  | CharData x -> x ^ scan_and_expand()
	  | PERef n ->
	      let en = dtd # par_entity n in
	      let (x,_) = en # replacement_text in
	      x ^ scan_and_expand()
	  | Eof ->
	      ""
	  | _ ->
	      assert false
      in
      let rtext = scan_and_expand() in
      resolver # close_in;
      resolver_is_open <- false;
      rtext, true
	(* TODO:
	 * - The replaced text is not parsed [VALIDATION WEAKNESS]
	 *)
  end
;;


class document_entity  the_resolver the_dtd the_name the_warner the_ext_id
                       init_errors_with_line_numbers
  =
  object (self)
    inherit external_entity  the_resolver the_dtd the_name the_warner
                             the_ext_id false init_errors_with_line_numbers

    (* A document entity is an external entity that does not allow
     * conditional sections, and that forces that internal parameter entities
     * are properly nested.
     *)

    initializer
    allow_conditionals <- false;
    force_parameter_entity_parsing <- true;
    check_text_declaration <- false;
  end
;;


class internal_entity the_dtd the_name the_warner the_literal_value
                      the_p_internal_subset init_errors_with_line_numbers
  =
  (* An internal entity uses a "literal entity value" as character source.
   * This value is first expanded and preprocessed, i.e. character and
   * parameter references are expanded.
   *
   * 'the_p_internal_subset': indicates that the entity is declared in the
   * internal subset. Such entity declarations are not allowed to contain
   * references to parameter entities.
   *)

  object (self)
    inherit entity
              the_dtd the_name the_warner init_errors_with_line_numbers
	    as super

    val p_internal_subset = the_p_internal_subset

    val mutable replacement_text = ""
    val mutable contains_external_references = false
    val mutable p_parsed_actually = false
    val mutable is_open = false
    val mutable state = At_beginning


    initializer
    let lexbuf = Lexing.from_string the_literal_value in
    let rec scan_and_expand () =
      match Markup_lexers.scan_dtd_string lexbuf with
	  ERef n -> "&" ^ n ^ ";" ^ scan_and_expand()
	| CRef(-1) -> "\r\n" ^ scan_and_expand()
	| CRef(-2) -> "\r" ^ scan_and_expand()
	| CRef(-3) -> "\n" ^ scan_and_expand()
	| CRef k -> character warner k ^ scan_and_expand()
	| CharData x -> x ^ scan_and_expand()
	| PERef n ->
	    if p_internal_subset then
	      raise(WF_error("Restriction of the internal subset: parameter entity not allowed here"));
	    let en = dtd # par_entity n in
	    let (x, extref) = en # replacement_text in
	    contains_external_references <-
	      contains_external_references or extref;
	    x ^ scan_and_expand()
	| Eof ->
	    ""
	| _ ->
	    assert false
    in
    is_open <- true;
    replacement_text <- scan_and_expand();
    is_open <- false;
    normalize_newline <- false;

    method private set_encoding (e:string) =
      raise(Validation_error("The encoding cannot be changed in internal entities"))


    method open_entity force_parsing init_lex_id =
      if is_open then
	raise(Validation_error("Recursive reference to entity `" ^ name ^ "'"));

      p_parsed_actually <- force_parsing;
      lexbuf  <- Lexing.from_string replacement_text;
      prolog  <- None;
      accept_prolog <- false;
      lex_id  <- init_lex_id;
      delegated_entity <- None;
      cond_level <- 0;
      state <- if p_parsed_actually then At_beginning else In_stream;
      is_open <- true;
      line <- 1;
      column <- 0;
      pos <- 0;
      space_seen <- false;
      last_token <- Eof;

    method next_token =
      if not is_open then
	failwith ("Internal entity " ^ name ^ " not open");
      begin match state with
	  At_beginning ->
	    let tok = super # next_token in
	    (* if tok = Eof then begin
	         state <- At_end;
	         Eof
	        end else
	    *)
	    begin
	      state <- First_token tok;
	      Begin_entity
	    end
	| First_token tok ->
	      ( match tok with
		    Eof -> state <- At_end; End_entity
		  | _   -> state <- In_stream; tok
	      )
	| In_stream ->
	    let tok = super # next_token in
	    ( match tok with
		  Eof -> if p_parsed_actually then begin
			   state <- At_end; End_entity
		         end
		         else Eof
		| _   -> tok
	    )
	| At_end -> Eof
      end

    method close_entity =
      if not is_open then
	failwith ("Internal entity " ^ name ^ " not open");
      is_open <- false;
      lex_id

    method replacement_text =
      if is_open then
	raise(Validation_error("Recursive reference to entity `" ^ name ^ "'"));
      replacement_text, contains_external_references
  end
;;

(* ======================================================================
 * History:
 *
 * $Log: markup_entity.ml,v $
 * Revision 1.19  2000/05/01 15:18:44  gerd
 * 	Improved CRLF handling in the replacement text of entities.
 * 	Changed one error message.
 *
 * Revision 1.18  2000/04/30 18:18:39  gerd
 * 	Bugfixes: The conversion of CR and CRLF to LF is now hopefully
 * done right. The new variable "normalize_newline" indicates whether
 * normalization must happen for that type of entity. The normalization
 * if actually carried out separately for every token that needs it.
 *
 * Revision 1.17  2000/03/13 23:42:38  gerd
 * 	Removed the resolver classes, and put them into their
 * own module (Markup_reader).
 *
 * Revision 1.16  2000/02/22 01:06:58  gerd
 * 	Bugfix: Resolvers are properly re-initialized. This bug caused
 * that entities could not be referenced twice in the same document.
 *
 * Revision 1.15  2000/01/20 20:54:11  gerd
 * 	New config.errors_with_line_numbers.
 *
 * Revision 1.14  2000/01/08 18:59:03  gerd
 * 	Corrected the string resolver.
 *
 * Revision 1.13  1999/09/01 22:58:23  gerd
 * 	Method warn_not_latin1 raises Illegal_character if the character
 * does not match the Char production.
 * 	External entities that are not document entities check if the
 * <?xml...?> declaration at the beginning matches the TextDecl production.
 * 	Method xml_declaration has type ... list option, not ... list.
 * 	Tag_beg and Tag_end now carry an entity_id with them.
 * 	The code to check empty entities has changed. That the Begin_entity/
 * End_entity pair is not to be added must be explicitly turned on. See the
 * description of empty entity handling in design.txt.
 * 	In internal subsets entity declarations are not allowed to refer
 * to parameter entities. The internal_entity class can do this now.
 * 	The p_parsed parameter of internal_entity has gone. It was simply
 * superflous.
 *
 * Revision 1.12  1999/09/01 16:24:13  gerd
 * 	The method replacement_text returns the text as described for
 * "included in literal". The former behaviour has been dropped to include
 * a leading and a trailing space character for parameter entities.
 * 	Bugfix: When general entities are included, they are always parsed.
 *
 * Revision 1.11  1999/08/31 19:13:31  gerd
 * 	Added checks on proper PE nesting. The idea is that tokens such
 * as Decl_element and Decl_rangle carry an entity ID with them. This ID
 * is simply an object of type < >, i.e. you can only test on identity.
 * The lexer always produces tokens with a dummy ID because it does not
 * know which entity is the current one. The entity layer replaces the dummy
 * ID with the actual ID. The parser checks that the IDs of pairs such as
 * Decl_element and Decl_rangle are the same; otherwise a Validation_error
 * is produced.
 *
 * Revision 1.10  1999/08/19 01:06:41  gerd
 * 	Improved error messages: external entities print their
 * ext id, too
 *
 * Revision 1.9  1999/08/15 20:35:48  gerd
 * 	Improved error messages.
 * 	Before the tokens Plus, Star, Qmark space is not allowed any longer.
 * 	Detection of recursive entity references is a bit cleaner.
 *
 * Revision 1.8  1999/08/15 15:33:44  gerd
 * 	Revised whitespace checking: At certain positions there must be
 * white space. These checks cannot be part of the lexer, as %entity; counts
 * as white space. They cannot be part of the yacc parser because one look-ahead
 * token would not suffice if we did that. So these checks must be done by the
 * entity layer. Luckily, the rules are simple: There are simply a number of
 * token pairs between which white space must occur independently of where
 * these token have been found. Two variables, "space_seen", and "last_token"
 * have been added in order to check these rules.
 *
 * Revision 1.7  1999/08/15 00:41:06  gerd
 * 	The [ token of conditional sections is now allowed to occur
 * in a different entity.
 *
 * Revision 1.6  1999/08/15 00:29:02  gerd
 * 	The method "attlist_replacement_text" has gone. There is now a
 * more general "replacement_text" method that computes the replacement
 * text for both internal and external entities. Additionally, this method
 * returns whether references to external entities have been resolved;
 * this is checked in the cases where formerly "attlist_replacement_text"
 * was used as it is not allowed everywhere.
 * 	Entities have a new slot "need_spaces" that indicates that the
 * next token must be white space or a parameter reference. The problem
 * was that "<!ATTLIST%e;" is legal because when including parameter
 * entities white space is added implicitly. Formerly, the white space
 * was expected by the underlying lexer; now the lexer does not check
 * anymore that "<!ATTLIST" is followed by white space because the lexer
 * cannot handle parameter references. Because of this, the check on
 * white space must be done by the entity.
 *
 * Revision 1.5  1999/08/14 22:57:19  gerd
 * 	It is allowed that external entities are empty because the
 * empty string is well-parsed for both declarations and contents. Empty
 * entities can be referenced anywhere because the references are replaced
 * by nothing. Because of this, the Begin_entity...End_entity brace is only
 * inserted if the entity is non-empty. (Otherwise references to empty
 * entities would not be allowed anywhere.)
 * 	As a consequence, the grammar has been changed such that a
 * single Eof is equivalent to Begin_entity,End_entity without content.
 *
 * Revision 1.4  1999/08/14 22:11:19  gerd
 *         Several objects have now a "warner" as argument which is
 * an object with a "warn" method. This is used to warn about characters
 * that cannot be represented in the Latin 1 alphabet.
 * 	Previously, the resolvers had features in order to warn about
 * such characters; this has been removed.
 * 	UTF-8 streams can be read even if they contain characters
 * that cannot be represented by 16 bits.
 * 	The buffering used in the resolvers is now solved in a
 * cleaner way; the number of characters that are expected to be read
 * from a source can be limited. This removes a bug with UTF-16 streams
 * that previously lead to wrong exceptions; and the buffering is more
 * efficient, too.
 *
 * Revision 1.3  1999/08/11 14:58:53  gerd
 * 	Some more names for encodings are allowed, such as "utf8" instead
 * of the standard name "UTF-8".
 * 	'resolve_as_file' interprets relative file names as relative to
 * the "parent" resolver.
 *
 * Revision 1.2  1999/08/10 21:35:07  gerd
 * 	The XML/encoding declaration at the beginning of entities is
 * evaluated. In particular, entities have now a method "xml_declaration"
 * which returns the name/value pairs of such a declaration. The "encoding"
 * setting is interpreted by the entity itself; "version", and "standalone"
 * are interpreted by Markup_yacc.parse_document_entity. Other settings
 * are ignored (this does not conform to the standard; the standard prescribes
 * that "version" MUST be given in the declaration of document; "standalone"
 * and "encoding" CAN be declared; no other settings are allowed).
 * 	TODO: The user should be warned if the standard is not exactly
 * fulfilled. -- The "standalone" property is not checked yet.
 *
 * Revision 1.1  1999/08/10 00:35:51  gerd
 * 	Initial revision.
 *
 *
 *)
