(* $Id: markup_lexers.mll,v 1.14 2000/05/01 15:19:23 gerd Exp $
 * ----------------------------------------------------------------------
 * Markup! The validating XML parser for Objective Caml.
 * Copyright 1999 by Gerd Stolpmann. See LICENSE for details.
 *)

{
  open Markup_types
  open Markup_lexer_types

  let get_name_end s k =
    (* Get the index of the end+1 of the name beginning at position k *)
    let l = String.length s in
    let rec find j =
      if j < l then
	match s.[j] with
	    ('A'..'Z'|
	     'a'..'z'|
	     '\192'..'\214'|
	     '\216'..'\246'|
	     '\248'..'\255'|
	     '0'..'9'|
	     '.'|
	     ':'|
	     '-'|
	     '_'|
	     '\183')   ->   find (j+1)
	  | _          ->   j
      else
	l
    in
    find k

  let get_ws_end s k =
    let l =  String.length s in
    let rec find j =
      if j < l then
	match s.[j] with
	    (' '|'\t'|'\r'|'\n') -> find (j+1)
	  | _                    -> j
      else
	l
    in
    find k

  let scan_pi lexbuf xml_scanner =
    let s = Lexing.lexeme lexbuf in
    let i = get_name_end s 2 in
    let j = get_ws_end s i in
    let s_name = String.sub s 2 (i-2) in
    let s_param = String.sub s j (String.length s - 2 - j) in
    (* prerr_endline s_name; *)
    (* prerr_endline s_param; *)
    if i=j & s_param <> "" then
      raise (WF_error ("Bad processing instruction"));
    match s_name with
	"xml" -> begin
	  let xml_lexbuf = Lexing.from_string (s_param ^ " ") in
	  let rec collect () =
	    let t = xml_scanner xml_lexbuf in
	    (* prerr_endline (string_of_int (Lexing.lexeme_end xml_lexbuf)); *)
	    if t = Pro_eof then
	      []
	    else
	      t :: collect()
	  in
	  PI_xml (collect())
	end
      | _ -> PI(s_name,s_param)

  class dummy_entity = object end

  let dummy_entity = ( new dummy_entity : entity_id )

  (* The following tokens are pre-allocated to reduce the load on the
   * GC.
   *)

  let tok_Doctype__Document_type = Doctype dummy_entity, Document_type
  let tok_Comment__Document      = Comment, Document
  let tok_Comment__Content       = Comment, Content
  let tok_Comment__Declaration   = Comment, Declaration
  let tok_Ignore__Document       = Ignore, Document
  let tok_Ignore__Within_tag     = Ignore, Within_tag
  let tok_Ignore__Document_type  = Ignore, Document_type
  let tok_Ignore__Declaration    = Ignore, Declaration
  let tok_Ignore__Ignored        = Ignore, Ignored_section
  let tok_Eof__Document          = Eof, Document
  let tok_Eof__Content           = Eof, Content
  let tok_Eof__Within_tag        = Eof, Within_tag
  let tok_Eof__Document_type     = Eof, Document_type
  let tok_Eof__Declaration       = Eof, Declaration
  let tok_Eof__Ignored           = Eof, Ignored_section
  let tok_LineEndCRLF__Content   = LineEnd "\r\n", Content
  let tok_LineEndCR__Content     = LineEnd "\r", Content
  let tok_LineEndLF__Content     = LineEnd "\n", Content
  let tok_CharDataRBRACKET__Content = CharData "]", Content
  let tok_Eq__Within_tag         = Eq, Within_tag
  let tok_Rangle__Content        = Rangle, Content
  let tok_Rangle_empty__Content  = Rangle_empty, Content
  let tok_Dtd_begin__Declaration = Dtd_begin dummy_entity, Declaration
  let tok_Doctype_rangle__Document = Doctype_rangle dummy_entity, Document
  let tok_Percent__Declaration   = Percent, Declaration
  let tok_Plus__Declaration      = Plus, Declaration
  let tok_Star__Declaration      = Star, Declaration
  let tok_Bar__Declaration       = Bar, Declaration
  let tok_Comma__Declaration     = Comma, Declaration
  let tok_Qmark__Declaration     = Qmark, Declaration
  let tok_Lparen__Declaration    = Lparen dummy_entity, Declaration
  let tok_RparenPlus__Declaration   = RparenPlus dummy_entity, Declaration
  let tok_RparenStar__Declaration   = RparenStar dummy_entity, Declaration
  let tok_RparenQmark__Declaration  = RparenQmark dummy_entity, Declaration
  let tok_Rparen__Declaration    = Rparen dummy_entity, Declaration
  let tok_Required__Declaration  = Required, Declaration
  let tok_Implied__Declaration   = Implied, Declaration
  let tok_Fixed__Declaration     = Fixed, Declaration
  let tok_Pcdata__Declaration    = Pcdata, Declaration
  let tok_Decl_element__Declaration  = Decl_element dummy_entity, Declaration
  let tok_Decl_attlist__Declaration  = Decl_attlist dummy_entity, Declaration
  let tok_Decl_entity__Declaration   = Decl_entity dummy_entity, Declaration
  let tok_Decl_notation__Declaration = Decl_notation dummy_entity, Declaration
  let tok_Conditional_begin__Declaration = Conditional_begin, Declaration 
  let tok_Conditional_begin__Ignored     = Conditional_begin, Ignored_section
  let tok_Conditional_end__Declaration   = Conditional_end, Declaration
  let tok_Conditional_end__Ignored       = Conditional_end, Ignored_section
  let tok_Conditional_body__Declaration  = Conditional_body, Declaration
  let tok_Decl_rangle__Declaration   = Decl_rangle dummy_entity, Declaration
  let tok_Dtd_end__Document_type     = Dtd_end dummy_entity, Document_type
}

let ws = [ ' ' '\t' '\r' '\n' ]

(* Note: ISO-8859-1 charset does not have 'combining characters' *)

let letter = ['A'-'Z' 'a'-'z' '\192'-'\214' '\216'-'\246' '\248'-'\255']
let extender = '\183'
let digit = ['0'-'9']
let hexdigit = ['0'-'9' 'A'-'F' 'a'-'f']
let namechar = letter | digit | '.' | ':' | '-' | '_' | extender
let name = ( letter | '_' | ':' ) namechar*
let nmtoken = namechar+

let pi_string = [^ '?' '\000'-'\008' '\011'-'\012' '\014'-'\031']* 
		( '?' [^ '>' '\000'-'\008' '\011'-'\012' '\014'-'\031'] 
                      [^ '?' '\000'-'\008' '\011'-'\012' '\014'-'\031']* )* 
                '?'?
let comment_string = [^ '-' '\000'-'\008' '\011'-'\012' '\014'-'\031' ]* 
                     ('-' [^ '-' '\000'-'\008' '\011'-'\012' '\014'-'\031' ]+ )*

(* let character = [^ '&' '<' '\000'-'\008' '\011'-'\012' '\014'-'\031' ] *)
let normal_character = 
  [^ '&' '<' ']' '\000'-'\008' '\010'-'\031']

let no_rbracket = [^ ']' '\000'-'\008' '\011'-'\012' '\014'-'\031']
let no_rbracket_no_rangle = [^ ']' '>'
                               '\000'-'\008' '\011'-'\012' '\014'-'\031']

let cdata_string = no_rbracket* ( "]" no_rbracket+ |
                                  "]]" ']'* no_rbracket_no_rangle no_rbracket*
				)*
                                ']'*
(* cdata_string = char* - ( char* ']]>' char* ) *)




(* scan_document: Lexer for the outermost structures *)

rule scan_document = parse
    "<?" name ws* pi_string "?>"
      { scan_pi lexbuf scan_xml_pi, Document }
  | "<?"
      { raise (WF_error ("Illegal processing instruction")) }
  | "<!DOCTYPE"
      { tok_Doctype__Document_type }
  | "<!--" comment_string "-->"
      { tok_Comment__Document }
  | "<!--"
      { raise (WF_error ("Illegal comment; possible reasons are illegal control characters or double hyphens ('--') inside the comment, or missing '-->'"))
      }
  | "<!"
      { raise (WF_error "Declaration either malformed or not allowed in this context") 
      }
  | "<" name
      { let s = Lexing.lexeme lexbuf in
	Tag_beg (String.sub s 1 (String.length s - 1), dummy_entity), Within_tag
      }
  | '<'
      { raise (WF_error ("Illegal token")) }
  | ws+
      { tok_Ignore__Document }
  | eof
      { tok_Eof__Document }
  | _
      { raise (WF_error ("Content not allowed here")) }

and scan_content = parse
    "<?" name ws* pi_string "?>"
      { scan_pi lexbuf scan_xml_pi, Content }
  | "<?"
      { raise (WF_error ("Illegal processing instruction")) }
  | "<!--" comment_string "-->"
      { tok_Comment__Content }
  | "<!--"
      { raise (WF_error ("Illegal comment; possible reasons are illegal control characters or double hyphens ('--') inside the comment, or missing '-->'"))
      }
  | "<" name
      { let s = Lexing.lexeme lexbuf in
	Tag_beg (String.sub s 1 (String.length s - 1), dummy_entity), 
	Within_tag 
      }
  | "</" name
      { let s = Lexing.lexeme lexbuf in
	Tag_end (String.sub s 2 (String.length s - 2), dummy_entity), 
	Within_tag 
      }
  | "<![CDATA[" cdata_string "]]>"
      { let s = Lexing.lexeme lexbuf in
	Cdata (String.sub s 9 (String.length s - 12)), Content }
  | "<!"
      { raise (WF_error "Declaration either malformed or not allowed in this context") 
      }
  | "<"
      { raise (WF_error ("The left angle bracket '<' must be written as '&lt;'"))
      }
  | "&#" digit+ ";"
      { let s = Lexing.lexeme lexbuf in
	CRef (int_of_string (String.sub s 2 (String.length s - 3))), Content }
  | "&#x" hexdigit+ ";"
      { let s = Lexing.lexeme lexbuf in
	CRef (int_of_string ("0x" ^ String.sub s 3 (String.length s - 4))), Content }
  | "&" name ";"
      { let s = Lexing.lexeme lexbuf in
	ERef (String.sub s 1 (String.length s - 2)), Content }
  | "&" 
      { raise (WF_error ("The ampersand '&' must be written as '&amp;'"))
      }

  (* LineEnd: Depending on whether we are reading from a primary source
   * (file) or from the replacement text of an internal entity, line endings
   * must be normalized (converted to \n) or not.
   * The entity classes do that. The yacc parser will never see LineEnd;
   * this token is always converted to the appropriate CharData token.
   *)

  | '\013' '\010'
      { tok_LineEndCRLF__Content }
  | '\013'
      { tok_LineEndCR__Content }
  | '\010'
      { tok_LineEndLF__Content }
  | eof
      { tok_Eof__Content }
  | "]]>" 
      { raise (WF_error ("The sequence ']]>' must be written as ']]&gt;'"))
      }
  | "]"
      { tok_CharDataRBRACKET__Content }
  | normal_character+
      { let s = Lexing.lexeme lexbuf in
	CharData s, Content 
      }
  | _
      { raise (Illegal_character 0) }


(* scan_within_tag_beg: after "<name" until matching ">" *)

and scan_within_tag = parse
    ws+
      { tok_Ignore__Within_tag }
  | name
      { Name (Lexing.lexeme lexbuf ), Within_tag }
  | '='
      { tok_Eq__Within_tag }
  | '"' [^ '"' ]* '"'
      { let s = Lexing.lexeme lexbuf in
	let v = String.sub s 1 (String.length s - 2) in
	Attval v, Within_tag }
  | '"'
      { raise (WF_error ("Cannot find the second quotation mark"))
      }
  | "'" [^ '\'' ]* "'"
      { let s = Lexing.lexeme lexbuf in
	let v = String.sub s 1 (String.length s - 2) in
	Attval v, Within_tag }
  | "'"
      { raise (WF_error ("Cannot find the second quotation mark"))
      }
  | '>'
      { tok_Rangle__Content }
  | "/>"
      { tok_Rangle_empty__Content }
  | eof
      { tok_Eof__Within_tag }
  | _
      { raise (WF_error ("Illegal inside tags")) }


(* scan_document_type: after "<!DOCTYPE" until matching ">" *)

and scan_document_type = parse
    name
      { let s = Lexing.lexeme lexbuf in
	Name s, Document_type }
  | ws+
      { tok_Ignore__Document_type }
  | '"' [^ '"']* '"'
      { let s = Lexing.lexeme lexbuf in
	(Unparsed_string (String.sub s 1 (String.length s - 2))), Document_type }
  | '"'
      { raise (WF_error ("Cannot find the second quotation mark"))
      }
  | "'" [^ '\'']* "'"
      { let s = Lexing.lexeme lexbuf in
	(Unparsed_string (String.sub s 1 (String.length s - 2))), Document_type }
  | "'"
      { raise (WF_error ("Cannot find the second quotation mark"))
      }
  | '['
      { tok_Dtd_begin__Declaration }
  | '>'
      { tok_Doctype_rangle__Document }
  | eof
      { tok_Eof__Document_type }
  | '&' 
      { raise (WF_error("References to general entities not allowed here")) }
  | '%' 
      { raise (WF_error("References to parameter entities not allowed here")) }
  | _
      { raise (WF_error("Content not allowed here")) }
	  


(* scan_declaration: after "[" in DTD until matching "]" *)

(* Note: The entity layer checks additionally that after some tokens must
 * be white space: Percent, Decl_*.
 * This must be done by the entity layer because a parameter reference
 * %name; implicitly inserts white space.
 *)

and scan_declaration = parse
    ws+
      { tok_Ignore__Declaration }
  | '%' name ';'
      { let s = Lexing.lexeme lexbuf in
	(PERef (String.sub s 1 (String.length s - 2))), Declaration }
  | '%'
      { tok_Percent__Declaration }
  | '&' 
      { raise(WF_error("References to general entities not allowed in DTDs")) }
  | name
      { Name (Lexing.lexeme lexbuf), Declaration }
  | nmtoken
      { Nametoken (Lexing.lexeme lexbuf), Declaration }
  | '+'
      { tok_Plus__Declaration }
  | '*'
      { tok_Star__Declaration }
  | '|'
      { tok_Bar__Declaration }
  | ','
      { tok_Comma__Declaration }
  | '?'
      { tok_Qmark__Declaration }
  | '('
      { tok_Lparen__Declaration }
  | ")+" 
      { tok_RparenPlus__Declaration }
  | ")*" 
      { tok_RparenStar__Declaration }
  | ")?"
      { tok_RparenQmark__Declaration }
  | ')'
      { tok_Rparen__Declaration }
  | "#" name
      { let s = Lexing.lexeme lexbuf in
	( match s with
	      "#REQUIRED" -> tok_Required__Declaration
	    | "#IMPLIED"  -> tok_Implied__Declaration
	    | "#FIXED"    -> tok_Fixed__Declaration
	    | "#PCDATA"   -> tok_Pcdata__Declaration
	    | _           -> raise (WF_error("Illegal token: " ^ s))
	)
      }
  | "<!ELEMENT"
      { tok_Decl_element__Declaration }
  | "<!ATTLIST"
      { tok_Decl_attlist__Declaration }
  | "<!ENTITY"
      { tok_Decl_entity__Declaration }
  | "<!NOTATION"
      { tok_Decl_notation__Declaration }
  | "<!--" comment_string "-->"
      { tok_Comment__Declaration }
  | "<!--"
      { raise (WF_error ("Illegal comment; possible reasons are illegal control characters or double hyphens ('--') inside the comment, or missing '-->'"))
      }
  | "<!["
      { tok_Conditional_begin__Declaration }
  | "]]>"
      { tok_Conditional_end__Declaration }
  | "["
      { tok_Conditional_body__Declaration }
  | "<?" name ws* pi_string "?>"
      { scan_pi lexbuf scan_xml_pi, Declaration }
  | "<?"
      { raise (WF_error ("Illegal processing instruction")) }
  | '"' [^ '"']* '"'
      { let s = Lexing.lexeme lexbuf in
	(Unparsed_string (String.sub s 1 (String.length s - 2))), Declaration }
  | '"'
      { raise (WF_error ("Cannot find the second quotation mark"))
      }
  | "'" [^ '\'']* "'"
      { let s = Lexing.lexeme lexbuf in
	(Unparsed_string (String.sub s 1 (String.length s - 2))), Declaration }
  | "'"
      { raise (WF_error ("Cannot find the second quotation mark"))
      }
  | '>'
      { tok_Decl_rangle__Declaration }
  | ']'
      { tok_Dtd_end__Document_type }
  | eof
      { tok_Eof__Declaration }
  | "<!"
      { raise (WF_error "Declaration either malformed or not allowed in this context") 
      }
  | _
      { raise (WF_error("Illegal token or character")) }


and scan_ignored_section = parse
  | "<!["
      { tok_Conditional_begin__Ignored }
  | "]]>"
      { tok_Conditional_end__Ignored }
  | eof
      { tok_Eof__Ignored }
  | [^ '<' '!' ']']+
      { tok_Ignore__Ignored }
  | _
      { tok_Ignore__Ignored }


and scan_xml_pi = parse
    name ws*
      { let s = Lexing.lexeme lexbuf in
	let j = get_name_end s 0 in
	Pro_name (String.sub s 0 j)
      }
  | "=" ws*
      { Pro_eq }
  | "'" [^ '\'']* "'" ws+
      { let s = Lexing.lexeme lexbuf in
	let j = String.index_from s 1 '\'' in
	Pro_string (String.sub s 1 (j-1))
      }
  | "'"
      { raise (WF_error ("Cannot find the second quotation mark"))
      }
  | '"' [^ '"']* '"' ws+
      { let s = Lexing.lexeme lexbuf in
	let j = String.index_from s 1 '"' in
	Pro_string (String.sub s 1 (j-1))
      }
  | '"'
      { raise (WF_error ("Cannot find the second quotation mark"))
      }
  | eof
      { Pro_eof }
  | _
      { (* prerr_endline (Lexing.lexeme lexbuf); *)
	raise (WF_error("Illegal token or character")) 
      }


(* The following scanner is used to determine the replacement text of
 * internal entities:
 *)

and scan_dtd_string = parse
    '%' name ';'
      { let s = Lexing.lexeme lexbuf in
	PERef (String.sub s 1 (String.length s - 2)) }
  | '%'
      { raise(WF_error("The character '%' must be written as '&#37;'")) }
  | '&' name ';'
      { let s = Lexing.lexeme lexbuf in
	ERef (String.sub s 1 (String.length s - 2)) }
  | "&#" digit+ ";"
      { let s = Lexing.lexeme lexbuf in
	CRef (int_of_string (String.sub s 2 (String.length s - 3))) }
  | "&#x" hexdigit+ ";"
      { let s = Lexing.lexeme lexbuf in
	CRef (int_of_string ("0x" ^ String.sub s 3 (String.length s - 4))) }
  | '&'
      { raise(WF_error("The character '&' must be written as '&amp;'")) }
  | '\013' '\010'
      { CRef(-1) }
  | '\013'
      { CRef(-2) }
  | '\010'
      { CRef(-3) }
  | [^ '&' '%' '\000'-'\008' '\011'-'\031' ]+
      { CharData (Lexing.lexeme lexbuf) }
  | eof
      { Eof }
  | _
      { raise (Illegal_character 0) }


(* This lexer is used to expand and normalize attribute values: *)

and scan_content_string = parse
    '&' name ';'
      { let s = Lexing.lexeme lexbuf in
	ERef (String.sub s 1 (String.length s - 2)) }
  | "&#" digit+ ";"
      { let s = Lexing.lexeme lexbuf in
	CRef (int_of_string (String.sub s 2 (String.length s - 3))) }
  | "&#x" hexdigit+ ";"
      { let s = Lexing.lexeme lexbuf in
	CRef (int_of_string ("0x" ^ String.sub s 3 (String.length s - 4))) }
  | '&'
      { raise(WF_error("The character '&' must be written as '&amp;'")) }
  | [^ '&' '\000'-'\031' '<' ]+
      { CharData (Lexing.lexeme lexbuf) }
  | '\009'
      { CRef 32 }
  | '\013' '\010'
      { CRef(-1)   (* A special case *)
      }
  | '\013'
      { CRef 32 }
  | '\010'
      { CRef 32 }
  | '<'
      { 
	(* Depending on the situation, '<' may be legal or not: *)
	CharData "<" 
      }
  | eof
      { Eof }
  | _
      { raise (Illegal_character 0) }


and scan_name_string = parse
    name
      { Name (Lexing.lexeme lexbuf) }
  | ws+
      { Ignore }
  | nmtoken
      { Nametoken (Lexing.lexeme lexbuf) }
  | eof
      { Eof }
  | _
      { CharData (Lexing.lexeme lexbuf) }

and scan_only_xml_decl = parse
    "<?xml" ws+ pi_string "?>"
      { scan_pi lexbuf scan_xml_pi }
  | ""
      { Eof }

and scan_for_crlf = parse
  | '\013' '\010'
      { CharData "\n" }
  | '\013'
      { CharData "\n" }
  | '\010'
      { CharData "\n" }
  | [^ '\010' '\013' ]+
      { CharData (Lexing.lexeme lexbuf) }
  | eof 
      { Eof }

(* ======================================================================
 * History:
 *
 * $Log: markup_lexers.mll,v $
 * Revision 1.14  2000/05/01 15:19:23  gerd
 * 	Improved CRLF handling in definitions of internal entities.
 * 	Changed error messages.
 *
 * Revision 1.13  2000/04/30 18:20:41  gerd
 * 	New lexer scan_for_crlf.
 * 	Corrected the lexers for correct CR and CRLF interpretation.
 * 	Improvement: New tok_* variables to avoid that constant
 * token pairs are allocated every time again.
 *
 * Revision 1.12  2000/04/28 01:50:07  gerd
 * 	Bug: It was possible that CR characters were not recognized.
 *
 * Revision 1.11  1999/09/01 23:00:03  gerd
 * 	Tag_beg and Tag_end have now an entity_id parameter.
 * 	<?xml a="x"b="y"?> is now rejected because of the missing space
 * between the attributes.
 * 	Added some more error conditions to improve the quality of
 * error messages.
 *
 * Revision 1.10  1999/09/01 16:24:57  gerd
 * 	Improved the quality of error messages.
 *
 * Revision 1.9  1999/08/31 19:13:31  gerd
 * 	Added checks on proper PE nesting. The idea is that tokens such
 * as Decl_element and Decl_rangle carry an entity ID with them. This ID
 * is simply an object of type < >, i.e. you can only test on identity.
 * The lexer always produces tokens with a dummy ID because it does not
 * know which entity is the current one. The entity layer replaces the dummy
 * ID with the actual ID. The parser checks that the IDs of pairs such as
 * Decl_element and Decl_rangle are the same; otherwise a Validation_error
 * is produced.
 *
 * Revision 1.8  1999/08/19 22:21:00  gerd
 * 	Bugfix: '>' in CDATA is allowed.
 *
 * Revision 1.7  1999/08/15 15:34:26  gerd
 * 	After "<!DOCTYPE", no white space is expected anymore (these
 * checks have been moved to the entity layer).
 *
 * Revision 1.6  1999/08/15 13:47:48  gerd
 * 	Processing instructions and comments are not allowed to contain
 * control characters.
 *
 * Revision 1.5  1999/08/15 02:22:09  gerd
 * 	Bugfixes: "]]>" is not allowed as character data even outside
 * CDATA sections.
 * 	'<' is now always a separate CharData token in attribute
 * values (the "expand_attvalue" function relies on this).
 *
 * Revision 1.4  1999/08/15 00:40:22  gerd
 * 	Some rules no longer demand that the keyword is followed by
 * white space. This is now done by the entity layer.
 * 	Additional lexer scan_only_xml_decl that detects either
 * <?xml ...?> or the empty string.
 *
 * Revision 1.3  1999/08/14 22:14:16  gerd
 * 	Bugfix: The token for "processing instructions" in Content
 * environment does no longer switch to Document environment. Instead,
 * Content environment remains.
 *
 * Revision 1.2  1999/08/10 21:35:10  gerd
 * 	The XML/encoding declaration at the beginning of entities is
 * evaluated. In particular, entities have now a method "xml_declaration"
 * which returns the name/value pairs of such a declaration. The "encoding"
 * setting is interpreted by the entity itself; "version", and "standalone"
 * are interpreted by Markup_yacc.parse_document_entity. Other settings
 * are ignored (this does not conform to the standard; the standard prescribes
 * that "version" MUST be given in the declaration of document; "standalone"
 * and "encoding" CAN be declared; no other settings are allowed).
 * 	TODO: The user should be warned if the standard is not exactly
 * fulfilled. -- The "standalone" property is not checked yet.
 *
 * Revision 1.1  1999/08/10 00:35:51  gerd
 * 	Initial revision.
 *
 *
 *)
