(* $Id: markup_reader.ml,v 1.1 2000/03/13 23:41:44 gerd Exp $
 * ----------------------------------------------------------------------
 * Markup! The validating XML parser for Objective Caml.
 * Copyright by Gerd Stolpmann. See LICENSE for details.
 *)

open Markup_types;;

class type resolver =
  object
    method open_in : ext_id -> Lexing.lexbuf
    method close_in : unit
    method change_encoding : string -> unit
    method clone : resolver
  end
;;


class virtual resolve_general (the_warner : collect_warnings) =
  object (self)
    val mutable encoding = ""
    val mutable encoding_requested = false

    val warner = the_warner


    method clone =
      ( {< encoding = "";
	   encoding_requested = false;
	>}
	: # resolver :> resolver )


    method private warn_not_latin1 (k:int) =
      (* Called if a character not representable in Latin 1 has been found.
       * k is the character code.
       *)
	if k < 0xd800 or (k >= 0xe000 & k <= 0xfffd) or
	   (k >= 0x10000 & k <= 0x10ffff) then begin
	     warner # warn ("Character not in Latin 1 alphabet: " ^ string_of_int k);
	   end
	else
	  raise (Illegal_character 0)

    method private autodetect s =
      (* s must be at least 4 bytes long. The slot 'encoding' is
       * set to:
       * "UTF-16-BE": UTF-16/UCS-2 encoding big endian
       * "UTF-16-LE": UTF-16/UCS-2 encoding little endian
       * "UTF-8":     UTF-8 encoding
       *)
      if String.length s < 4 then
	encoding <- "UTF-8"
      else if String.sub s 0 2 = "\254\255" then
	encoding <- "UTF-16-BE"
      else if String.sub s 0 2 = "\255\254" then
	encoding <- "UTF-16-LE"
      else
	encoding <- "UTF-8"

    method private utf8_to_latin1 s_in l_in s_out l_out_max =
      (* Parameters see 'to_latin1' *)
      let l = l_in in
      let s = s_in in
      let u = s_out in
      let rec recode j k =
	if j < l & k < l_out_max then begin
	  match s.[j] with
	      ('\000'..'\127' as c) ->
		u.[k] <- c;
		recode (j+1) (k+1)
	    | ('\128'..'\223' as c) ->
		if j+1 >= l then
		  k, j
		else begin
		  let n1 = Char.code c in
		  let n2 = Char.code (s.[j+1]) in
		  if n2 < 128 or n2 > 191 then raise Bad_character_stream;
		  let n = ((n1 land 0b11111) lsl 6) lor (n2 land 0b111111) in
		  if n > 255 then
		    self # warn_not_latin1 n
		  else
		    u.[k] <- Char.chr n;
		  recode (j+2) k
		end
	    | ('\224'..'\239' as c) ->
		if j+2 >= l then
		  k, j
		else begin
		  let n1 = Char.code c in
		  let n2 = Char.code (s.[j+1]) in
		  let n3 = Char.code (s.[j+2]) in
		  if n2 < 128 or n2 > 191 then raise Bad_character_stream;
		  if n3 < 128 or n3 > 191 then raise Bad_character_stream;
		  let n = ((n1 land 0b1111) lsl 12) lor
			  ((n2 land 0b111111) lsl 6) lor
			  (n3 land 0b111111) in
		  self # warn_not_latin1 n;
		  recode (j+3) k
		end
	    | ('\240'..'\247' as c) ->
		if j+3 >= l then
		  k, j
		else begin
		  let n1 = Char.code c in
		  let n2 = Char.code (s.[j+1]) in
		  let n3 = Char.code (s.[j+2]) in
		  let n4 = Char.code (s.[j+3]) in
		  if n2 < 128 or n2 > 191 then raise Bad_character_stream;
		  if n3 < 128 or n3 > 191 then raise Bad_character_stream;
		  if n4 < 128 or n4 > 191 then raise Bad_character_stream;
		  let n = ((n1 land 0b111) lsl 18) lor
			  ((n2 land 0b111111) lsl 12) lor
			  ((n3 land 0b111111) lsl 6) lor
			  (n4 land 0b111111) in
		  self # warn_not_latin1 n;
		  recode (j+4) k
		end
	    | ('\248'..'\251' as c) ->
		if j+4 >= l then
		  k, j
		else begin
		  let n1 = Char.code c in
		  let n2 = Char.code (s.[j+1]) in
		  let n3 = Char.code (s.[j+2]) in
		  let n4 = Char.code (s.[j+3]) in
		  let n5 = Char.code (s.[j+4]) in
		  if n2 < 128 or n2 > 191 then raise Bad_character_stream;
		  if n3 < 128 or n3 > 191 then raise Bad_character_stream;
		  if n4 < 128 or n4 > 191 then raise Bad_character_stream;
		  if n5 < 128 or n5 > 191 then raise Bad_character_stream;
		  let n = ((n1 land 0b11) lsl 24) lor
			  ((n2 land 0b111111) lsl 18) lor
			  ((n3 land 0b111111) lsl 12) lor
			  ((n4 land 0b111111) lsl 6) lor
			  (n5 land 0b111111) in
		  self # warn_not_latin1 n;
		  recode (j+5) k
		end
	    | ('\252'..'\253' as c) ->
		if j+5 >= l then
		  k, j
		else begin
		  let n1 = Char.code c in
		  let n2 = Char.code (s.[j+1]) in
		  let n3 = Char.code (s.[j+2]) in
		  let n4 = Char.code (s.[j+3]) in
		  let n5 = Char.code (s.[j+4]) in
		  let n6 = Char.code (s.[j+5]) in
		  if n2 < 128 or n2 > 191 then raise Bad_character_stream;
		  if n3 < 128 or n3 > 191 then raise Bad_character_stream;
		  if n4 < 128 or n4 > 191 then raise Bad_character_stream;
		  if n5 < 128 or n5 > 191 then raise Bad_character_stream;
		  if n6 < 128 or n6 > 191 then raise Bad_character_stream;
		  let n = ((n1 land 0b11) lsl 30) lor
			  ((n2 land 0b111111) lsl 24) lor
			  ((n3 land 0b111111) lsl 18) lor
			  ((n4 land 0b111111) lsl 12) lor
			  ((n5 land 0b111111) lsl 6) lor
			  (n6 land 0b111111) in
		  self # warn_not_latin1 n;
		  recode (j+6) k
		end
	    | _ ->
		raise Bad_character_stream
	end
	else
	  k, j
      in
      recode 0 0


    method private utf16be_to_latin1 s_in l_in s_out l_out_max =
      (* Parameters see 'to_latin1' *)
      let l = l_in in
      let s = s_in in
      let u = s_out in
      let rec recode j k =
	if j < l-1 & k < l_out_max then begin
	  let n = (Char.code(s.[j]) lsl 8) lor (Char.code(s.[j+1])) in
	  if n < 256 then begin
	    u.[k] <- Char.chr n;
	    recode (j+2) (k+1)
	  end
	  else begin
	    match n with
		0xfeff ->
		  (* big endian byte order mark *)
		  encoding <- "UTF-16-BE";
		  recode (j+2) k
	      | 0xfffe ->
		  (* little endian byte order mark *)
		  encoding <- "UTF-16-LE";
		  let s_in' = String.sub s j (l-j) in
		  let l_in' = l-j in
		  let l_out_max' = l_out_max - k in
		  let s_out' = String.create l_out_max' in
		  let l_out, n_consumed =
		    self # utf16le_to_latin1 s_in' l_in' s_out' l_out_max' in
		  String.blit s_out' 0 s_out k l_out;
		  k+l_out, j+n_consumed
	      | _ ->
		  self # warn_not_latin1 n;
		  recode (j+2) k
	  end
	end
	else
	  k, j
      in
      recode 0 0


    method private utf16le_to_latin1 s_in l_in s_out l_out_max =
      (* Parameters see 'to_latin1' *)
      let l = l_in in
      let s = s_in in
      let u = s_out in
      let rec recode j k =
	if j < l-1 & k < l_out_max then begin
	  let n = (Char.code(s.[j+1]) lsl 8) lor (Char.code(s.[j])) in
	  if n < 256 then begin
	    u.[k] <- Char.chr n;
	    recode (j+2) (k+1)
	  end
	  else begin
	    match n with
		0xfffe ->
		  (* big endian byte order mark *)
		  encoding <- "UTF-16-BE";
		  let s_in' = String.sub s j (l-j) in
		  let l_in' = l-j in
		  let l_out_max' = l_out_max - k in
		  let s_out' = String.create l_out_max' in
		  let l_out, n_consumed =
		    self # utf16be_to_latin1 s_in' l_in' s_out' l_out_max' in
		  String.blit s_out' 0 s_out k l_out;
		  k+l_out, j+n_consumed
	      | 0xfeff ->
		  (* little endian byte order mark *)
		  encoding <- "UTF-16-LE";
		  recode (j+2) k
	      | _ ->
		  self # warn_not_latin1 n;
		  recode (j+2) k
	  end
	end
	else
	  k, j
      in
      recode 0 0


    method private to_latin1 s_in l_in s_out l_out_max =
      (* The string 's_in' is converted to the Latin 1 string 's_out'.
       * Only the first 'l_in' bytes of 's_in' are converted.
       * At most 'l_out_max' characters are written to 's_out'.
       * The method returns (l_out_act,n_consumed). 'l_out_act' are the number
       * of characters written to 's_out'. 'n_consumed' are the number of
       * characters of 's_in' that have actually been converted. '(n_consumed'
       * may be lower than 'l_in' because of incomplete character sequences.)
       *
       * Possible encodings of 's_in' as determined by 'encoding':
       * "UTF-16", "UTF-16-BE", "UTF-16-LE", "UTF-8",
       * "UCS-2", "ISO-10646-UCS-2",
       * "ISO-8859-1".
       * 'encoding' is changed into a normalized value.
       *)
      match String.uppercase encoding with
	  ("UTF-16"|"UTF16"|"ISO-10646-UCS-2") ->
	    encoding <- "UTF-16";
	    self # utf16be_to_latin1 s_in l_in s_out l_out_max
	| "UTF-16-BE" ->
	    self # utf16be_to_latin1 s_in l_in s_out l_out_max
	| "UTF-16-LE" ->
	    self # utf16le_to_latin1 s_in l_in s_out l_out_max
	| ("UTF-8"|"UTF8") ->
	    encoding <- "UTF-8";
	    self # utf8_to_latin1 s_in l_in s_out l_out_max
	| ("ISO-8859-1"|"ISO88591"|"ISO8859-1"|"ISO-88591"|"") ->
	    encoding <- "ISO-8859-1";
	    let n = min l_in l_out_max in
	    String.blit s_in 0 s_out 0 n;
	    n, n
	| _ ->
	    raise (WF_error "Unsupported character encoding")


    method private virtual next_string : string -> int -> int -> int
    method private virtual init_in : ext_id -> unit
    method virtual close_in : unit

    method open_in xid =
      let buffer_max = 512 in
      let buffer = String.make buffer_max ' ' in
      let buffer_len = ref 0 in
      let buffer_end = ref false in
      let fillup () =
	if not !buffer_end & !buffer_len < buffer_max then begin
	  let l =
	    self # next_string buffer !buffer_len (buffer_max - !buffer_len) in
	  if l = 0 then
	    buffer_end := true
	  else begin
	    buffer_len := !buffer_len + l
	  end
	end
      in
      let consume n =
	let l = !buffer_len - n in
	String.blit buffer n buffer 0 l;
	buffer_len := l
      in

      encoding <- "";
      encoding_requested <- false;
      self # init_in xid;
      fillup();
      self # autodetect buffer;

      Lexing.from_function
	(fun s n ->
	   fillup();
	   if !buffer_len = 0 then
	     0
	   else begin
	     let m_in  = !buffer_len in
	     let m_out = if encoding_requested then n else 1 in
	     let l_out,n_consumed =
	       self # to_latin1 buffer m_in s m_out in
	     if n_consumed = 0 then
	       raise Bad_character_stream;
	       (* failwith "Badly encoded character"; *)
	     consume n_consumed;
	     l_out
	   end)

    method change_encoding enc =
      if List.mem (String.uppercase enc)
	   [ "ISO-8859-1"; "UTF-16"; "ISO-10646-UCS-2";
	     "UTF-16-BE"; "UTF-16-LE"; "UTF-8";
	     "UTF16"; "UTF8"; "ISO88591"; "ISO8859-1";
	     "ISO-88591" ] then begin
	encoding <- enc;
	encoding_requested <- true;
      end
      else if enc = "" then
	encoding_requested <- true
      else
	raise (WF_error("Unsupported character encoding"))

  end
;;


class resolve_read_channel ch the_warner =
  object (self)
    inherit resolve_general the_warner

    (* This resolver takes the in_channel ch as character source. *)
    val ch = (ch : in_channel)
    method private init_in (_:ext_id) = ()
    method private next_string s ofs len =
      input ch s ofs len
    method close_in = ()
    method clone =
      failwith "This resolver for external references cannot open this entity"

  end
;;


class resolve_read_string str =
  object (self)
    inherit resolve_general (new collect_warnings)

    (* This resolver takes the string str as character source. *)
    val str = (str : string)
    val mutable pos = 0
    method private init_in (_:ext_id) = ()
    method private next_string s ofs len =
      let l = min len (String.length str - pos) in
      String.blit str pos s ofs l;
      pos <- pos + l;
      l
    method close_in = ()
    method clone =
      failwith "This resolver for external references cannot open this entity"
  end
;;


class resolve_as_file the_warner =
  object (self)
    inherit resolve_general the_warner as super

    (* This resolver interprets file names given as system identifiers. *)
    (* TODO:
     * - The file: prefix should be recognized
     * - How are relative paths interpreted?
     *)
    val mutable ch = stdin
    val mutable directory = "."

    method clone =
      ( {< encoding = "";
	   encoding_requested = false;
	   ch = stdin;
	>}
	: #resolver :> resolver
      )

    method private init_in xid =
      let open_file fname =
	let fname' =
	  if Filename.is_relative fname then
	    Filename.concat directory fname
	  else
	    fname
	in
	open_in fname', Filename.dirname fname'
      in

      let the_ch, the_directory =
	match xid with
	    System fname     -> open_file fname
	  | Public (_,fname) -> open_file fname
      in
      ch <- the_ch;
      directory <- the_directory

    method private next_string s ofs len =
      input ch s ofs len

    method close_in =
      close_in ch

  end
;;


(* ======================================================================
 * History:
 * 
 * $Log: markup_reader.ml,v $
 * Revision 1.1  2000/03/13 23:41:44  gerd
 * 	Initial revision; this code was formerly part of Markup_entity.
 *
 * 
 *)
