/* $Id: markup_yacc.mly,v 1.19 2000/05/01 15:20:08 gerd Exp $
 * ----------------------------------------------------------------------
 * Markup! The validating XML parser for Objective Caml.
 * Copyright by Gerd Stolpmann. See LICENSE for details.
 */


%{
  *)

  open Parsing

  open Markup_types

  open Markup_lexer_types
    (* This 'open' overrides the definition of the type 'token' that has been
     * generated by ocamlyacc.
     * THIS IS VERY CRITICAL.
     *)

  open Markup_dtd
  open Markup_entity
  open Markup_document
  open Markup_aux

  (* Some types from the interface definition: *)

  type config =
      { warner : collect_warnings;
	errors_with_line_numbers : bool;
	processing_instructions_inline : bool;
	virtual_root : bool;
	sgml_capitalize_names : bool;   (* TODO *)
	sgml_attribute_values : bool;   (* TODO *)
	sgml_omit_endtags : bool;       (* TODO *)
	debugging_mode : bool;
      }

  type source =
      Entity of ((dtd -> Markup_entity.entity) * Markup_reader.resolver)
    | Channel of in_channel
    | File of string
    | Latin1 of string
    | ExtID of (ext_id * Markup_reader.resolver)


  type 'ext domspec =
      { map : (node_type, 'ext node) Hashtbl.t;
	default_element : 'ext node;
      }
      (* specifies which node to use as exemplar for which node type *)



  (* This parser works with side effects. This is not optimal but currently
   * the simplest approach. In order to have better control of the side
   * effects, the parser is artifically turned into an object. This is done
   * by a "sed" script in the Makefile that changes some of the generated "let"
   * definitions into "val" definitions and the other "let" definitions into
   * "method" definitions.
   * We begin here the class:
   *)

  (**********************************************************************)

  class ['ext] parser_object
      init_doc init_dtd init_extend_dtd init_config init_resolver init_dom =
    object (self)

      (* Note that the 'ext parameter has been the motivation to make the
       * parser a class.
       *)

      val dtd = init_dtd
	(* The DTD being parsed; or the DTD currently assumed *)

      val extend_dtd = init_extend_dtd
	(* Whether the DTD should be extended by ELEMENT, ATTLIST, and
	 * NOTATION declarations or not.
	 *)

      val doc = init_doc
        (* The current document *)

      method doc = (doc : 'ext document)

      val resolver = init_resolver
        (* The resolver for external IDs *)

      val config = init_config
        (* The current configuration *)

      val elstack = (Stack.create() : ('ext node * entity_id) Stack.t)
       (* The element stack containing all open elements, i.e. elements that
	* have begun by a start tag but that have not been finished (end tag).
	* If the parser sees a start tag, it creates the element and pushes it
	* on top of this stack. If the parser recognizes an end tag, it pulls
	* one element from the stack and checks if it has the same name as
	* given with the end tag.
	*
	* At initialization time, a special element is pushed on the stack,
	* the so-called virtual root (vr). It is always the bottommost
	* element of the stack, and serves as a guard.
	* [See "initializer" below.]
	*)

      method current =
	(* Get the top element of the element stack *)
	(* In O'Caml 3: Stack.top *)
	try
	  let x = Stack.pop elstack in
	  Stack.push x elstack;
	  let el, entid = x in
	  el
	with
	    Stack.Empty -> assert false
		(* Not possible, because the VR is always the element at the
		 * bottom of the stack.
		 *)

      val mutable n_tags_open = 0
	(* Number of begin tags that have been parsed and whose corresponding
	 * end tags have not yet been parsed
	 *)

      val mutable p_internal_subset = false
        (* true while parsing the internal subset *)

      val mutable p_internal_subset_stack = Stack.create()
        (* Saves the value of 'p_internal_subset' while external entities
	 * are being parsed. This is a crude hack that needs some cleanup.
	 *)

      val mutable root = None
        (* Contains the root element (topmost element) while it is being parsed
	 * and after it has been parsed.
	 * This variable is None before the root element is seen.
	 *)

      method root = root

      val dom = init_dom
        (* A hashtable that contains exemplar objects for the various element
	 * types. If an element is parsed, the exemplar is looked up and
	 * "cloned" (by the "create" method)
	 *)

      val data_exemplar =
	try
	  Hashtbl.find init_dom.map T_data
	with
	    Not_found ->
	      failwith "Markup_yacc.parser_object: The initial DOM map does not specify T_data nodes"

      method only_whitespace data =
	(* Checks that the string "data" contains only whitespace. On failure,
	 * Validation_error is raised.
	 *)
	let lexbuf = Lexing.from_string data in
	let t1 = Markup_lexers.scan_name_string lexbuf in
	if t1 <> Ignore then
	  raise(WF_error("Data not allowed here"));
	let t2 = Markup_lexers.scan_name_string lexbuf in
	if t2 <> Eof then
	  raise(WF_error("Data not allowed here"));
	()

      initializer
	(* --- Initialize 'elstack': Push the VR on the stack. *)
	(* If we are NOT in virtual_root mode, any element serves as VR,
	 * because no method is ever called on this element. We simply
	 * take the default exemplar.
	 * IF we are in virtual_root mode, the VR is created (almost)
	 * normally.
	 *)
        let vr =
	  if config.virtual_root then begin
	    (* Get the 'exemplar', as usual: *)
            let exemplar =
	      try Hashtbl.find dom.map (T_element "-pi")
	      with Not_found -> dom.default_element in
	    (* A trick: while the 'vr' is being created, the DTD is
	     * changed such that this is allowed.
	     *)
	    let aa = dtd # arbitrary_allowed in        (* save old state    *)
	    dtd # allow_arbitrary;
	    let vroot =
	      exemplar # create_element dtd (T_element "-vr") [] in
	    if not aa then dtd # disallow_arbitrary;   (* restore old state *)
	    vroot
	  end
	  else
	    dom.default_element
	in
	(* Move it to the stack: *)
	Stack.push (vr, (self :> entity_id)) elstack;

%}

/* See markup_types.ml for comments */

/* The following definition of 'token' MUST be identical to the definition
 * given in Markup_types.
 * Note: If the definitions do not match, the error is not always reported;
 * it may result in segfaults (it's a hack...)
 */

%token Begin_entity
%token End_entity
%token Comment
%token Ignore
%token Eq
%token Rangle
%token Rangle_empty
%token Conditional_begin
%token Conditional_body
%token Conditional_end
%token Percent
%token Plus
%token Star
%token Bar
%token Comma
%token Qmark
%token Pcdata
%token Required
%token Implied
%token Fixed
%token Eof

%token <entity_id> Doctype
%token <entity_id> Doctype_rangle
%token <entity_id> Dtd_begin
%token <entity_id> Dtd_end
%token <entity_id> Decl_element
%token <entity_id> Decl_attlist
%token <entity_id> Decl_entity
%token <entity_id> Decl_notation
%token <entity_id> Decl_rangle
%token <entity_id> Lparen
%token <entity_id> Rparen
%token <entity_id> RparenPlus
%token <entity_id> RparenStar
%token <entity_id> RparenQmark

%token <string*entity_id>  Tag_beg
%token <string*entity_id>  Tag_end

%token <string*string>     PI
%token <prolog_token list> PI_xml
%token <string>            Cdata
%token <int>               CRef
%token <string>            ERef
%token <string>            PERef
%token <string>            CharData
%token <string>            LineEnd
%token <string>            Name
%token <string>            Nametoken
%token <string>            Attval
%token <string>            Attval_nl_normalized
%token <string>            Unparsed_string

/* START SYMBOLS:
 *
 * "ext_document":       parses a complete XML document (i.e. containing a
 *                       <!DOCTYPE..> and an element)
 * "ext_declarations":   parses an "external DTD subset", i.e. a sequence
 *                       of declarations
 * "ext_element":        parses a single element (no <!DOCTYPE...> allowed);
 *                       the element needs not to be the root element of the
 *                       DTD
 *
 * The functions corresponding to these symbols return always () because
 * they only have side-effects.
 */

%start ext_document
%start ext_declarations
%start ext_element

%type <unit> ext_document
%type <unit> ext_declarations
%type <unit> ext_element

/* SOME GENERAL COMMENTS:
 *
 * The parser does not get its tokens from the lexers directly. Instead of
 * this, there is an entity object between the parser and the lexers. This
 * object already handles:
 *
 * - References to general and parameter entities. The token stream is
 *   modified such that tokens automatically come from the referenced entities.
 *   External parameter entities and all general entities are embraced by
 *   the two special tokens Begin_entity and End_entity. The parser must
 *   check that these braces are correctly nested.
 *   TODO: The nesting conditions of internal parameter entities are currently
 *         not checked.
 * - Conditional sections.
 * - The tokens Ignore and Comment are removed from the token stream.
 *   This means that the parser does not see white space outside elements.
 */

%%

/* Note: XML declarations <?xml ...?> are handled by the entity objects,
 * not by the grammar.
 */

ext_document:
  Begin_entity misc_then_prolog_then_rest
    {
      if n_tags_open <> 0 then
	raise(WF_error("Missing end tag"))
    }

misc_then_prolog_then_rest:
  misc misc_then_prolog_then_rest
    { () }
| doctypedecl misc_s contents_start End_entity
    { () }
| contents_start End_entity
    { () }

misc_s:
  /* sequence of white space or processing instructions */
  misc misc_s
    { () }
|
    { () }

misc:
  pi
    { () }
| CharData
    /* In this context, the lexers do not recognize white space; instead
     * CharData tokens containing white space are delivered.
     */
    { self # only_whitespace $1 }


ext_element:
  Begin_entity misc_s contents_start End_entity
    {
      if n_tags_open <> 0 then
	raise(WF_error("Missing end tag"))
    }

/********************* DOCUMENT TYPE DECLARATION *************************/

doctypedecl:
  /* parses from <!DOCTYPE to >. As side-effect, first the declarations of
   * the internal DTD (if any) are put into !!on_dtd, the the declarations
   * of the external DTD (if any) are put into this DTD object.
   * Note that "internal_dtd_opt" is always reduced before this rule, and
   * because of this the declarations of the internal DTD are parsed first.
   */
  Doctype Name external_id_opt internal_dtd_opt Doctype_rangle
    { let doctype_entid = $1 in
      let doctype_rangle_entid = $5 in
      let root_name = $2 in
      let external_subset = $3 in
      let internal_subset = $4 in
      if doctype_entid != doctype_rangle_entid then
	raise (Validation_error("Entities not properly nested with DOCTYPE declaration"));
      dtd # set_root root_name;
      begin match external_subset, internal_subset with
	  None, false    -> ()         (* no DTD means no ID *)
	| None, true     -> dtd # set_id Internal
	| Some id, false -> dtd # set_id (External id)
	| Some id, true  -> dtd # set_id (Derived id)
      end;
      (* Get now the external doctype declaration. Note that the internal
       * subset has precedence and must be read first.
       *)
      begin match $3 with
	  None -> ()
	| Some id ->
	    let r' = resolver # clone in
	    let pobj =
	      new parser_object
		(new document config.warner)
		dtd
		extend_dtd
		config
		r'
		dom
	    in
	    let en = new external_entity r' dtd "[dtd]"
		         config.warner id false config.errors_with_line_numbers
	    in
	    try
	      en # open_entity true Declaration;
	      pobj # ext_declarations
		(fun _ -> en # next_token)
		(Lexing.from_string "");
	      ignore(en # close_entity);
	    with
		error ->
		  ignore(en # close_entity);
		  let pos = en # position_string in
		  raise (At(pos, error))
      end;
      dtd # validate
    }

external_id_opt:
  /* parses SYSTEM "id" or PUBLIC "id" "id" or nothing. Returns
   * Some (System "id") or Some (Public "id") or None.
   */
  external_id
    { Some $1 }
|
    { None }


/* Note that there are no keywords for SYSTEM or PUBLIC, because these would
 * be difficult to recognize in the lexical contexts. We are lucky because
 * SYSTEM is followed by exectly one string while PUBLIC is followed by two
 * strings, so we can distinguish between them.
 */

external_id:
  Name Unparsed_string   /* where Name = SYSTEM */
    { if $1 = "SYSTEM" then
	System $2
      else
	raise(WF_error("SYSTEM expected"))
    }
| Name Unparsed_string Unparsed_string  /* where Name = PUBLIC */
    { if $1 = "PUBLIC" then begin
	check_public_id $2;
	Public($2,$3)
      end
      else
	raise(WF_error("PUBLIC expected"))
    }

internal_dtd_opt:
  /* parses the internal DTD subset from "[" to "]", if any. As side-effect,
   * the parsed declarations are put into the dtd object.
   */
  internal_dtd
    { true }
|
    { false }

internal_dtd:
  internal_dtd_begin declarations  internal_dtd_end
    { let dtd_begin_entid = $1 in
      let dtd_end_entid = $3 in
      if dtd_begin_entid != dtd_end_entid then
	raise(Validation_error("Entities not properly nested with internal DTD subset"))
    }


internal_dtd_begin:
  Dtd_begin
    { assert (not p_internal_subset);
      p_internal_subset <- true }

internal_dtd_end:
  Dtd_end
    { assert p_internal_subset;
      p_internal_subset <- false }


ext_declarations:
  /* Parses a sequence of declarations given by an entity. As side-effect,
   * the parsed declarations are put into the dtd object.
   */
  Begin_entity declarations End_entity
   { () }
| Eof
   { () }

declarations:
  /* Parses a sequence of declarations... Note that the declarations are
   * processed in the order they are written. (Because of this, "markupdecl"
   * must be a separate rule which is reduced before all the "declarations"
   * rules are reduced. If you would merge the rules, in the style of
   *
   * declarations:
   *   Decl_element Name contentspec Decl_rangle declarations
   * | Decl_attlist Name attdef_factor declarations
   * ...
   *
   * the declarations would be processed in the reverse order.)
   */
  markupdecl declarations
    { () }
| begin_entity declarations end_entity declarations
    { () }
|
    { () }
/* PERef already resolved by entity object */

begin_entity : Begin_entity
   { (* Set 'p_internal_subset' to 'false' until the matching 'end_entity'
      * rule is parsed. This allows unrestricted usage of parameter entities
      * within declarations of internal entities.
      *)
     Stack.push p_internal_subset p_internal_subset_stack;
     p_internal_subset <- false;
   }

end_entity : End_entity
   { (* Restore the old value of 'p_internal_subset'. *)
     p_internal_subset <- Stack.pop p_internal_subset_stack;
   }

markupdecl:
  /* Parses a single declaration (or processing instruction). As side-effect
   * the parsed declaration is entered into the dtd object.
   */
  elementdecl
    { () }
| attlistdecl
    { () }
| entitydecl
    { () }
| notationdecl
    { () }
| PI
    { let target, value = $1 in
      let pi = new proc_instruction target value in
      dtd # add_pinstr pi
    }


/*************************** ELEMENT DECLARATIONS ********************/

elementdecl:
  /* parses <!ELEMENT ... >. Puts the parsed element as side-effect into
   * dtd.
   */
  Decl_element Name contentspec Decl_rangle
    { let decl_element_entid = $1 in
      let decl_rangle_entid = $4 in
      let name = $2 in
      if decl_element_entid != decl_rangle_entid then
	raise (Validation_error "Entities not properly nested with ELEMENT declaration");
      if extend_dtd then begin
	let el = new dtd_element (dtd) name in
	(* It is allowed that an <!ATTLIST...>  precedes the corresponding
	 * <!ELEMENT...>. Because of this it is possible that there is already
	 * an element called 'name' in the DTD, and we must set the content
	 * model of this element.
	 *)
	try
	  dtd # add_element el;
	  el # set_content_model $3;
	with
	    Not_found ->  (* means: there is already an element 'name' *)
	      let el' = dtd # element name in
	      el' # set_content_model $3;
              (* raises Validation_error if el' already has a content model *)
      end
    }
| Decl_element Begin_entity
    { raise (Validation_error("This entity reference is not allowed here")) }

/* Note that the lexers again cannot recognize keywords such as EMPTY or
 * ANY because these identifiers are only keywords if they occur at special
 * positions.
 */

contentspec:
  /* parses a content model and returns it (type content_model_type) */
  Name   /* EMPTY or ANY */
    { match $1 with
	  "EMPTY" -> Empty
	| "ANY"   -> Any
	| _       -> raise(WF_error("EMPTY, ANY, or a subexpression expected"))
    }
| choice_or_seq
    { Regexp $1 }
| mixed
    { $1 }



multiplier_opt:
  /* returns the multipler symbol (?,*,+) or Ignore if not found */
  multiplier
    { $1 }
|
    { Ignore }

multiplier:
  /* returns one of the multiplier symbols (?,*,+) */
  Plus
    { Plus }
| Star
    { Star }
| Qmark
    { Qmark }
| Begin_entity
    { raise (Validation_error("This entity reference is not allowed here")) }

/* TODO: Lparen - Rparen: entity_ids vergleichen */
mixed :
  Lparen Pcdata mixed_alternatives_top
    { let lparen_entid = $1 in
      let rest, rparen_entid = $3 in
      if lparen_entid != rparen_entid then
	raise (Validation_error "Entities not properly nested with parentheses");
      Mixed (MPCDATA :: rest)
    }

mixed_alternatives_top:
  Rparen
    { [], $1 }
| RparenStar
    { [], $1 }
| Bar mixed_alternatives
    { $2 }

mixed_alternatives :
  Name RparenStar
    { let name = $1 in
      let rparen_entid = $2 in
      [MChild name], rparen_entid
    }
| Name Bar mixed_alternatives
    { let name = $1 in
      let rest, rparen_entid = $3 in
      (MChild name :: rest), rparen_entid
    }

choice_or_seq:
  /* parses either a regular expression, or a mixed expression. Returns
   * Mixed spec or Regexp spec (content_model_type).
   * Which kind of expression (regexp or mixed) is being read is recognized
   * after the first subexpression has been parsed; the other subexpressions
   * must be of the same kind.
   */
  Lparen cp choice_or_seq_factor
    { let lparen_entid = $1 in
      let re = $2 in
      let (finalmark,subexpr), rparen_entid = $3 in
      if lparen_entid != rparen_entid then
	raise (Validation_error "Entities not properly nested with parentheses");
      (* Check that the other subexpressions are "regexp", too, and
       * merge them with the first.
       *)
      let re' =
	match subexpr with
	    Alt []  ->  re
	  | Alt alt -> Alt (re :: alt)
	  | Seq seq -> Seq (re :: seq)
	  | _       -> assert false
      in
      (* Interpret the finalmark. *)
      match finalmark with
	  Ignore -> re'
	| Plus   -> Repeated1 re'
	| Star   -> Repeated re'
	| Qmark  -> Optional re'
	| _      -> assert false
    }
| Begin_entity
    { raise (Validation_error("This entity reference is not allowed here")) }

choice_or_seq_factor:
  /* Parses "|<subexpr>|...)" or ",<subexpr>,...)", both forms optionally
   * followed by ?, *, or +.
   * Returns ((finalmark, expr), rparen_entid), where
   * - finalmark is the character after the right parenthesis or Ignore
   * - expr is either
   *   Alt []              meaning that only ")" has been found
   *   Alt non_empty_list  meaning that the subexpressions are separated by '|'
   *   Seq non_empty_list  meaning that the subexpressions are separated by ','
   */
  Rparen
    { (Ignore, Alt []), $1 }
| RparenPlus
    { (Plus, Alt []), $1 }
| RparenStar
    { (Star, Alt []), $1 }
| RparenQmark
    { (Qmark, Alt []), $1 }
| Bar cp choice_or_seq_factor
    { let re = $2 in
      let (finalmark, subexpr), rparen_entid = $3 in
      begin match subexpr with
	  Alt []  -> (finalmark, (Alt [re])), rparen_entid
	| Alt alt -> (finalmark, (Alt (re :: alt))), rparen_entid
	| _       -> raise(WF_error("It is not allowed to mix alternatives and sequences"))
      end
    }
| Comma cp choice_or_seq_factor
    { let re = $2 in
      let (finalmark, subexpr), rparen_entid = $3 in
      begin match subexpr with
	  Alt []  -> (finalmark, (Seq [re])), rparen_entid
	| Seq seq -> (finalmark, (Seq (re :: seq))), rparen_entid
	| _       -> raise(WF_error("It is not allowed to mix alternatives and sequences"))
      end
    }
| Begin_entity
    { raise (Validation_error("This entity reference is not allowed here")) }

cp:
  /* parse either a name, or a parenthesized subexpression "(...)"  */
  Name multiplier_opt
    { match $2 with
	  Ignore -> Child $1
	| Plus   -> Repeated1 (Child $1)
	| Star   -> Repeated  (Child $1)
	| Qmark  -> Optional  (Child $1)
	| _      -> assert false
    }
| choice_or_seq
    { $1 }


/********************* ATTRIBUTE LIST DECLARATION ***********************/

attlistdecl:
  /* parses <!ATTLIST ... >. Enters the attribute list in dtd as side-
   * effect.
   */
  Decl_attlist Name attdef_factor
    { let decl_attlist_entid = $1 in
      let el_name = $2 in
      let at_list, decl_rangle_entid = $3 in
      if decl_attlist_entid != decl_rangle_entid then
	raise (Validation_error "Entities not properly nested with ATTLIST declaration");
      if extend_dtd then begin
	let new_el = new dtd_element dtd el_name in
	(* Note that it is allowed that <!ATTLIST...> precedes the corresponding
	 * <!ELEMENT...> declaration. In this case we add the element declaration
	 * already to the DTD but leave the content model unspecified.
	 *)
	let el =
	  try
	    dtd # add_element new_el;
	    new_el
	  with
	      Not_found ->  (* already added *)
		let old_el = dtd # element el_name in
		if old_el # attribute_names <>  [] then
		  config.warner # warn ("More than one ATTLIST declaration for element type `" ^
					el_name ^ "'");
		old_el
	in
	List.iter
	  (fun (a_name, a_type, a_default) ->
	     el # add_attribute a_name a_type a_default)
	  at_list
      end
    }
| Decl_attlist Begin_entity
    { raise (Validation_error("This entity reference is not allowed here")) }

attdef_factor:
  /* parses a list of triples <name> <type> <default value> and returns the
   * list as (string * att_type * att_default) list.
   */
  attdef attdef_factor
    { let attdef = $1 in
      let attdef_rest, decl_rangle_entid = $2 in
      (attdef :: attdef_rest), decl_rangle_entid }
| Decl_rangle
    { [], $1 }
| Begin_entity
    { raise (Validation_error("This entity reference is not allowed here")) }

attdef:
  /* Parses a single triple */
  Name atttype defaultdecl
    { ($1,$2,$3) }

/* In 'atttype' we again must recognize keywords. Again this is possible:
 * NOTATION is followed by a left parenthesis, and can be recognized by
 * the parser. The other keywords are grammatically equivalent (they do
 * not introduce their own phrases).
 * An enumeration can be recognized by a single left parenthesis. Note that
 * the lexer returns "nmtokens" that match the rules for "names" as Name,
 * not Nametoken.
 *
 * Note: The XML specification does not require that the entities are properly
 * nested with the parantheses of an attribute declaration.
 */


atttype:
  /* Parses an attribute type and returns it as att_type. */
  Name
    { match $1 with
	  "CDATA"    -> A_cdata
	| "ID"       -> A_id
	| "IDREF"    -> A_idref
	| "IDREFS"   -> A_idrefs
	| "ENTITY"   -> A_entity
	| "ENTITIES" -> A_entities
	| "NMTOKEN"  -> A_nmtoken
	| "NMTOKENS" -> A_nmtokens
	| _          -> raise(WF_error("One of CDATA, ID, IDREF, IDREFS, ENTITY, ENTITIES, NMTOKEN, NMTOKENS, NOTATION, or a subexpression expected"))
    }
| Name /* NOTATION */ Lparen Name notation_factor
    { if $1 = "NOTATION" then
	A_notation($3::$4)
      else
	raise(WF_error("NOTATION expected"))
    }
| Name Begin_entity
    { raise (Validation_error("This entity reference is not allowed here")) }
| Name Lparen Begin_entity
    { raise (Validation_error("This entity reference is not allowed here")) }
| Lparen Name nmtoken_factor           /* Enumeration */
    { A_enum($2 :: $3) }
| Lparen Nametoken nmtoken_factor
    { A_enum($2 :: $3) }
| Lparen Begin_entity
    { raise (Validation_error("This entity reference is not allowed here")) }
| Begin_entity
    { raise (Validation_error("This entity reference is not allowed here")) }

notation_factor:
  /* Parse "|<name>|...)" and return the name list */
  Bar Name notation_factor
    { $2 :: $3 }
| Rparen
    { [] }
| Begin_entity
    { raise (Validation_error("This entity reference is not allowed here")) }

nmtoken_factor:
  /* Parse "|<nmtoken>|...)" and return the nmtoken list */
  Bar Name nmtoken_factor
    { $2 :: $3 }
| Bar Nametoken nmtoken_factor
    { $2 :: $3 }
| Rparen
    { [] }
| Begin_entity
    { raise (Validation_error("This entity reference is not allowed here")) }


/* The default values must be expanded and normalized. This has been implemented
 * by the function expand_attvalue.
 */


defaultdecl:
  /* Parse the default value for an attribute and return it as att_default */
  Required
    { D_required }
| Implied
    { D_implied }
| Fixed Unparsed_string
    { D_fixed (expand_attvalue dtd $2 config.warner false) }
| Unparsed_string
    { D_default (expand_attvalue dtd $1 config.warner false) }
| Fixed Begin_entity
    { raise (Validation_error("This entity reference is not allowed here")) }
| Begin_entity
    { raise (Validation_error("This entity reference is not allowed here")) }


/**************************** ENTITY DECLARATION ***********************/

entitydecl:
  /* parses <!ENTITY ... >. The parsed entity declaration is entered into the
   * dtd object as side-effect.
   */
  Decl_entity Name entitydef Decl_rangle     /* A general entity */
    { let decl_entity_entid = $1 in
      let decl_rangle_entid = $4 in
      let name = $2 in
      if decl_entity_entid != decl_rangle_entid then
	raise (Validation_error "Entities not properly nested with ENTITY declaration");
      let en =
	(* Distinguish between
         * - internal entities
         * - external entities
         * - NDATA (unparsed) entities
         *)
	match $3 with
	    (Some s, None,     None)   ->
	      new internal_entity dtd name config.warner s p_internal_subset
		  config.errors_with_line_numbers
	  | (None,   Some xid, None)   ->
	      new external_entity (resolver # clone) dtd name config.warner
                                  xid false config.errors_with_line_numbers

	  | (None,   Some xid, Some n) ->
	      (new ndata_entity name xid n  :> entity)
	  | _ -> assert false
      in
      dtd # add_gen_entity en
    }
| Decl_entity Percent Name pedef Decl_rangle      /* A parameter entity */
    { let name = $3 in
      let en =
	(* Distinguish between internal and external entities *)
	match $4 with
	    (Some s, None)   ->
	      new internal_entity dtd name config.warner s p_internal_subset
  		  config.errors_with_line_numbers
	  | (None,   Some xid)   ->
	      new external_entity (resolver # clone) dtd name config.warner
                                  xid true config.errors_with_line_numbers
	  | _ -> assert false
      in
      dtd # add_par_entity en
    }
| Decl_entity Begin_entity
    { raise (Validation_error("This entity reference is not allowed here")) }
| Decl_entity Percent Begin_entity
    { raise (Validation_error("This entity reference is not allowed here")) }


entitydef:
  /* parses the definition value of a general entity. Returns either:
   * - (Some s, None,   None)    meaning the definition of an internal entity
   *                               with (literal) value s has been found
   * - (None,   Some x, None)    meaning that an external parsed entity with
   *                               external ID x has been found
   * - (None,   Some x, Some n)  meaning that an unparsed entity with
   *                               external ID x and notations n has been found
   */
  Unparsed_string
    { Some $1, None, None }
| external_id ndatadecl_opt
    {  None, Some $1, $2 }
| Begin_entity
    { raise (Validation_error("This entity reference is not allowed here")) }

pedef:
  /* parses the definition value of a parameter entity. Returns either:
   * - (Some s, None)     meaning that the definition of an internal entity
   *                        with (literal) value s has been found
   * - (None,   Some x)   meaning that an external ID x has been found
   */
  Unparsed_string
    { Some $1, None }
| external_id
    { None, Some $1 }
| Begin_entity
    { raise (Validation_error("This entity reference is not allowed here")) }

ndatadecl_opt:
  /* Parses either NDATA "string" or the empty string; returns Some "string"
   * in the former, None in the latter case.
   */
  Name Name
    { if $1 = "NDATA" then
	Some $2
      else
	raise(WF_error("NDATA expected"))
    }
|
    { None }
| Begin_entity
    { raise (Validation_error("This entity reference is not allowed here")) }


/**************************** NOTATION DECLARATION *******************/

notationdecl:
  /* parses <!NOTATION ... > and enters the notation declaration into the
   * dtd object as side-effect
   */
  Decl_notation Name Name /* SYSTEM or PUBLIC */ Unparsed_string string_opt
    Decl_rangle
    { let decl_notation_entid = $1 in
      let decl_rangle_entid = $6 in
      if decl_notation_entid != decl_rangle_entid then
	raise (Validation_error "Entities not properly nested with NOTATION declaration");
      let xid =
	(* Note that it is allowed that PUBLIC is only followed by one
	 * string literal
	 *)
	match $3 with
	    "SYSTEM" ->
	      if $5 <> None then raise(WF_error("SYSTEM must be followed only by one argument"));
	      System $4
	  | "PUBLIC" ->
	      begin match $5 with
		  None ->
		    check_public_id $4;
		    Public($4,"")
		| Some p ->
		    check_public_id $4;
		    Public($4,p)
	      end
	  | _ -> raise(WF_error("One of PUBLIC or SYSTEM expected"))
      in
      if extend_dtd then begin
	let no = new dtd_notation $2 xid in
	dtd # add_notation no
      end
    }
| Decl_notation Begin_entity
    { raise (Validation_error("This entity reference is not allowed here")) }
| Decl_notation Name Begin_entity
    { raise (Validation_error("This entity reference is not allowed here")) }
| Decl_notation Name Name Begin_entity
    { raise (Validation_error("This entity reference is not allowed here")) }
| Decl_notation Name Name Unparsed_string Begin_entity
    { raise (Validation_error("This entity reference is not allowed here")) }

string_opt:
  Unparsed_string
    { Some $1 }
|
    { None }

/************************* CONDITIONAL SECTION ***************************/

/* Conditional sections are implemented in the entity objects */


/****************************** ELEMENTS **************************/


contents_start:
  /* parses <element>...</element> misc*, i.e. exactly one element followed
   * optionally by white space or processing instructions.
   * The element is entered into the global variables as follows:
   * - If elstack is non-empty, the parsed element is added as new child to
   *   the top element of the stack.
   * - If elstack is empty, the root_examplar object is modified rather than
   *   that a new element is created. If additionally the variable root is
   *   None, it is assigned Some root_examplar.
   * Note that the modification of the root_exemplar is done by the method
   * internal_init.
   * The reason why the root element is modified rather than newly created
   * is a typing requirement. It must be possible that the class of the root
   * is derived from the original class element_impl, i.e. the user must be
   * able to add additional methods. If we created a new root object, we
   * would have to denote to which class the new object belongs; the root
   * would always be an 'element_impl' object (and not a derived object).
   * If we instead cloned an  exemplar object and modified it by the
   * "create" method, the root object would belong to the same class as the
   * exemplar (good), but the type of the parsing function would always
   * state that an 'element_impl' was created (because we can pass the new
   * object only back via a global variable). The only solution is to
   * modify the object that has been passed to the parsing function directly.
   */
  start_tag contents
    { () }

contents:
  /* parses a sequence of start tags, end tags, content, and processing
   * instructions. That the tags are properly nested is dynamically checked.
   * As result, recognized elements are added to their parent elements,
   * content is added to the element containing it, and processing instructions
   * are entered into the element embracing them. (All as side-effects.)
   */
  start_tag contents
    { () }
| end_tag contents
    { () }
| char_data contents
    { () }
| cref contents
    { () }
| pi contents
    { () }
| entity_ref contents
    { () }
|
    { () }

entity_ref:
   Begin_entity contents End_entity
    { if n_tags_open = 0 then
	raise(WF_error("Entity reference not allowed here"))
    }


start_tag:
  /* parses <element attribute-values> or <element attribute-values/>.
   *
   * EFFECT: If elstack is non-empty, the element is added to the
   * top element of the stack as new child, and the element
   * is pushed on the stack. If elstack is empty, the root_exemplar is
   * modified and gets the parsed name and attribute list. The root_exemplar
   * is pushed on the stack. If additionally the variable root is empty, too,
   * this variable is initialized.
   * If the <element ... /> form has been parsed, no element is pushed
   * on the stack.
   */
  Tag_beg attribute_list
    { 
      let emptiness, attlist = $2 in
      let name, tag_beg_entid = $1 in
      let exemplar =
	try Hashtbl.find dom.map (T_element name)
	with Not_found -> dom.default_element in
      let d =
	exemplar # create_element dtd (T_element name) attlist in

      if n_tags_open = 0 then begin
	if root = None then begin
	  (* We have found the begin tag of the root element. *)
	  if config.virtual_root then begin
	    (* The user wants the virtual root instead of the real root.
	     * The real root element becomes the child of the VR.
	     *)
	    (* Assertion: self # current is the virtual root *)
	    root <- Some (self # current);
	    self # current # add_node d;
	    doc # init_root (self # current);
	  end
	  else begin
	    (* Normal behaviour: The user wants to get the real root. *)
	    root <- Some d;
	    doc # init_root d;
	  end;
	end
	else
	  (* We have found a second topmost element. This is illegal. *)
	  raise(WF_error("Document must consist of only one toplevel element"))
      end
      else begin
	(* We have found some inner begin tag. *)
	self # current # add_node d
      end;

      if emptiness then
	(* An empty tag like <a/>. *)
	d # local_validate
      else begin
	(* A non-empty tag. *)
	Stack.push (d, tag_beg_entid) elstack;
	n_tags_open <- n_tags_open + 1;
      end;

    }

end_tag:
  /* parses </element>.
   * Pops the top element from the elstack and checks if it is the same
   * element.
   */
  Tag_end Rangle
    { let name, tag_end_entid = $1 in
      if n_tags_open = 0 then
	raise(WF_error("End-tag without start-tag"));
      let x, tag_beg_entid = Stack.pop elstack in
      let x_name =
	match x # node_type with
	    T_data -> assert false
	  | T_element n -> n
      in
      if name <> x_name then
	raise(WF_error("End-tag does not match start-tag"));
      if tag_beg_entid != tag_end_entid then
	raise(WF_error("End-tag not in the same entity as the start-tag"));
      x # local_validate;

      n_tags_open <- n_tags_open - 1;
      
      assert (n_tags_open >= 0);

    }

char_data:
  /* Parses any literal characters not otherwise matching, and adds the
   * characters to the top element of elstack.
   * If elstack is empty, it is assumed that there is no surrounding
   * element, and any non-white space character is forbidden.
   */
  CharData
    { let data = $1 in
      if n_tags_open = 0 then
	(* only white space is allowed *)
	self # only_whitespace data
      else
	self # current # add_node (data_exemplar # create_data dtd data)
    }
| Cdata
    { let data = $1 in
      if n_tags_open = 0 then
	raise (WF_error("CDATA section not allowed here"));
      self # current # add_node (data_exemplar # create_data dtd data)
    }

cref:
  /* Parses &#...; and adds the character to the top element of elstack. */
  CRef
    { let code = $1 in
      if n_tags_open = 0 then
	(* No surrounding element: character references are not allowed *)
	raise(WF_error("Character reference not allowed here"));
      self # current # add_node
	                 (data_exemplar # create_data
	                 dtd
	                 (character config.warner code))
    }

pi:
  /* Parses <?...?> (but not <?xml white-space ... ?>).
   * If there is a top element in elstack, the processing instruction is added
   * to this element.
   */
  PI
    { let target,value = $1 in
      if n_tags_open = 0 & not config.virtual_root
      then
	doc # add_pinstr (new proc_instruction target value)
      else begin
	(* Special case: if processing instructions are processed inline,
	 * they are wrapped into artificial "-pi" objects.
	 *)
	if config.processing_instructions_inline then begin
	  (* Get the 'exemplar', as usual: *)
	  let exemplar =
	    try Hashtbl.find dom.map (T_element "-pi")
	    with Not_found -> dom.default_element in
	  (* A trick: while the 'wrapper' is being created, the DTD is
	   * changed such that this is allowed.
	   *)
	  let aa = dtd # arbitrary_allowed in        (* save old state    *)
	  dtd # allow_arbitrary;
	  let wrapper =
	    exemplar # create_element dtd (T_element "-pi") [] in
	  if not aa then dtd # disallow_arbitrary;   (* restore old state *)
	  (* Put the PI into the 'wrapper' *)
	  wrapper # add_pinstr (new proc_instruction target value);
	  wrapper # local_validate;                  (* succeeds always   *)
	  self # current # add_node wrapper;
	  (* Note that a change in the function validate_content ensures
	   * that the parser does not complain about the additional elements
	   * in 'cur'
	   *)
	end
	else
	  (* Normal behaviour: Add the PI to the parent element. *)
	  self # current # add_pinstr (new proc_instruction target value)
      end
    }


attribute_list:
  /* Parses name="value" lists followed by ">" or "/>".
   * Returns a pair (empty, list):
   * - empty is a boolean and true iff the list is ended by "/>"
   * - list is the list of (name,value) pairs (both strings); the values
   *   are already expanded
   */
  Name Eq Attval attribute_list
    { let emptiness, attlist = $4 in
      emptiness, (($1, expand_attvalue dtd $3 config.warner true) :: attlist)
    }
| Name Eq Attval_nl_normalized attribute_list
    { let emptiness, attlist = $4 in
      emptiness, (($1, expand_attvalue dtd $3 config.warner false) :: attlist)
    }
| Rangle
    { false, [] }
| Rangle_empty
    { true, [] }

%%

(* Because ocamlyacc forgets adding a linenumber directive: *)

# 10000 "markup_yacc.mly"

(**********************************************************************)

(* Here ends the class definition: *)
end
;;

(**********************************************************************)

open Markup_reader;;


class default_ext =
  object(self)
    val mutable node = (None : ('a extension node as 'a) option)
    method clone = {< >}
    method node =
      match node with
	  None ->
	    assert false
	| Some n -> n
    method set_node n =
      node <- Some n
  end
;;


let default_extension = new default_ext;;

let default_dom =
  let d = Hashtbl.create 2 in
  Hashtbl.add d T_data (new data_impl default_extension "");
  { map = d;
    default_element = new element_impl default_extension
  }
;;

let parse_dtd_entity cfg src =
  (* Parse a DTD given as separate entity. *)
  let dtd = new dtd cfg.warner in
  let w = cfg.warner in
  let e = cfg.errors_with_line_numbers in
  let r, en =
    match src with
	Entity(m,r') -> r', m dtd
      | Channel ch -> let r = new resolve_read_channel ch w in
	              r,
		      new external_entity r dtd "[toplevel dtd]" w (System "")
			                  false e
      | File fname -> let r = new resolve_as_file w in
	              r,
		      new external_entity r dtd "[toplevel dtd]" w (System fname)
                                          false e
      | Latin1 str -> let r = new resolve_read_string str in
	              r,
		      new external_entity r dtd "[toplevel dtd]" w (System "")
                                          false e
      | ExtID(xid,r') -> 
  	              let r = r' # clone in
		      r, 
                      new external_entity r dtd "[toplevel dtd]" w xid false e
  in
  en # set_debugging_mode (cfg.debugging_mode);
  let pobj =
    new parser_object
      (new document cfg.warner)
      dtd
      true
      cfg
      r
      default_dom
  in
  try
    en # open_entity true Declaration;
    pobj # ext_declarations
      (fun _ -> en # next_token)
      (Lexing.from_string "");
    ignore(en # close_entity);
    dtd # validate;
    dtd
  with
      error ->
	ignore(en # close_entity);
	let pos = en # position_string in
	raise (At(pos, error))
;;


let parse_content_entity cfg src a_dtd dom_exempl =
  (* Parse an element given as separate entity *)
  a_dtd # validate;            (* ensure that the DTD is valid *)
  let w = cfg.warner in
  let e = cfg.errors_with_line_numbers in
  let r, en =
    match src with
	Entity(m,r') -> r', m a_dtd
      | Channel ch -> let r = new resolve_read_channel ch w in
	              r,
		      new external_entity r a_dtd "[toplevel element]" w
			                  (System "") false e
      | File fname -> let r = new resolve_as_file w in
	              r,
		      new external_entity r a_dtd "[toplevel element]" w
			                  (System fname) false e
      | Latin1 str -> let r = new resolve_read_string str in
	              r,
		      new external_entity r a_dtd "[toplevel element]" w
			                  (System "") false e
      | ExtID(xid,r') -> 
  	              let r = r' # clone in
		      r, 
		      new external_entity r a_dtd "[toplevel element]" w xid
                                          false e
  in
  en # set_debugging_mode (cfg.debugging_mode);
  let pobj =
    new parser_object
      (new document cfg.warner)
      a_dtd
      true
      cfg 
      r
      dom_exempl
  in
  try
    en # open_entity true Content;
    pobj # ext_element
      (fun _ ->
	 let tok = en # next_token in
	 tok)
      (Lexing.from_string "");
    ignore(en # close_entity);
    match pobj # root with
	Some r -> r
      | None -> raise(WF_error("No root element"))
  with
      error ->
	ignore(en # close_entity);
	let pos = en # position_string in
	raise (At(pos, error))
;;


let iparse_document_entity cfg src dom_exempl p_wf =
  (* Parse an element given as separate entity *)
  let a_dtd = new dtd cfg.warner in
  let e = cfg.errors_with_line_numbers in
  if p_wf then
    a_dtd # allow_arbitrary;
  let a_doc = new document cfg.warner in
  let w = cfg.warner in
  let r, en =
    match src with
	Entity(m,r') -> r', m a_dtd
      | Channel ch -> let r = new resolve_read_channel ch w in
	              r,
		      new document_entity r a_dtd "[toplevel document]" w (System "") e
      | File fname -> let r = new resolve_as_file w in
	              r,
		      new document_entity r a_dtd "[toplevel document]" w (System fname) e
      | Latin1 str -> let r = new resolve_read_string str in
	              r,
		      new document_entity r a_dtd "[toplevel document]" w (System "") e
      | ExtID(xid,r') ->
  	              let r = r' # clone in
		      r, 
		      new document_entity r a_dtd "[toplevel document]" w xid e
  in
  en # set_debugging_mode (cfg.debugging_mode);
  let pobj =
    new parser_object
      a_doc
      a_dtd
      (not p_wf)
      cfg 
      r
      dom_exempl
  in
  try
    en # open_entity true Document;          (* open the entity *)
    pobj # ext_document                      (* parse the entity *)
      (fun _ ->
	 let tok = en # next_token in
	 (* prerr_endline (string_of_tok tok); *)
	 tok)
      (Lexing.from_string "");
    let xml_decl = en # xml_declaration in   (* The <?xml ...?> instruction *)
    let d = pobj # doc in                    (* get the parsed document *)
    (* Interpret the <?xml ...?> instruction at the beginning of the document:*)
    begin match xml_decl with
	Some xml ->
	  let v, _, s = decode_doc_xml_pi xml in
	  check_version_num v;
	  d # init_xml_version v;
	  let v = match s with
	      None -> false
	    | Some "yes" -> true
	    | Some "no" -> false
	    | _ -> raise (WF_error("Illegal 'standalone' declaration"))
	  in
	  d # init_xml_standalone v
      | None ->
	  ()
    end;
    ignore(en # close_entity);               (* close the entity *)
    (* return the document to the caller *)
    d
  with
      error ->
	ignore(en # close_entity);
	let pos = en # position_string in
	raise (At(pos, error))
;;


let parse_document_entity cfg src dom_exempl =
  iparse_document_entity cfg src dom_exempl false;;
let parse_wf_entity cfg src dom_exempl =
  iparse_document_entity cfg src dom_exempl true;;


let default_config =
  let w = new collect_warnings in
  { warner = w;
    errors_with_line_numbers = true;
    processing_instructions_inline = false;
    virtual_root = false;
    sgml_capitalize_names = false;
    sgml_attribute_values = false;
    sgml_omit_endtags = false;
    debugging_mode = false;
  }


(* ======================================================================
 * History:
 *
 * $Log: markup_yacc.mly,v $
 * Revision 1.19  2000/05/01 15:20:08  gerd
 * 	"End tag matches start tag" is checked before "End tag in the
 * same entity as start tag".
 *
 * Revision 1.18  2000/04/30 18:23:08  gerd
 * 	Bigger change: Introduced the concept of virtual roots. First,
 * this reduces the number of checks. Second, it makes it possible to
 * return the virtual root to the caller instead of the real root (new
 * config options 'virtual_root' and 'processing_instructions_inline').
 * 	Minor changes because of better CR/CRLF handling.
 *
 * Revision 1.17  2000/03/13 23:47:46  gerd
 * 	Updated because of interface changes. (See markup_yacc_shadow.mli
 * rev. 1.8)
 *
 * Revision 1.16  2000/01/20 20:54:43  gerd
 * 	New config.errors_with_line_numbers.
 *
 * Revision 1.15  1999/12/17 22:27:58  gerd
 * 	Bugfix: The value of 'p_internal_subset' (an instance
 * variable of the parser object) is to true when the internal subset
 * begins, and is set to false when this subset ends. The error was
 * that references to external entities within this subset did not
 * set 'p_internal_subset' to false; this is now corrected by introducing
 * the 'p_internal_subset_stack'.
 * 	This is a typical example of how the code gets more and
 * more complicated and that it is very difficult to really understand
 * what is going on.
 *
 * Revision 1.14  1999/11/09 22:23:37  gerd
 * 	Removed the invocation of "init_dtd" of the root document.
 * This method is no longer available. The DTD is also passed to the
 * document object by the root element, so nothing essential changes.
 *
 * Revision 1.13  1999/10/25 23:37:09  gerd
 * 	Bugfix: The warning "More than one ATTLIST declaration for element
 * type ..." is only generated if an ATTLIST is found while there are already
 * attributes for the element.
 *
 * Revision 1.12  1999/09/01 23:08:38  gerd
 * 	New frontend function: parse_wf_document. This simply uses
 * a DTD that allows anything, and by the new parameter "extend_dtd" it is
 * avoided that element, attlist, and notation declarations are added to this
 * DTD. The idea is that this function simulates a well-formedness parser.
 * 	Tag_beg, Tag_end carry the entity_id. The "elstack" stores the
 * entity_id of the stacked tag. This was necessary because otherwise there
 * are some examples to produces incorrectly nested elements.
 * 	p_internal_subset is a variable that stores whether the internal
 * subset is being parsed. This is important beacause entity declarations in
 * internal subsets are not allowed to contain parameter references.
 * 	It is checked if the "elstack" is empty after all has been parsed.
 * 	Processing instructions outside DTDs and outside elements are now
 * added to the document.
 * 	The rules of mixed and regexp style content models have been
 * separated. The code is now much simpler.
 * 	Entity references outside elements are detected and rejected.
 *
 * Revision 1.11  1999/09/01 16:26:08  gerd
 * 	Improved the quality of error messages.
 *
 * Revision 1.10  1999/08/31 19:13:31  gerd
 * 	Added checks on proper PE nesting. The idea is that tokens such
 * as Decl_element and Decl_rangle carry an entity ID with them. This ID
 * is simply an object of type < >, i.e. you can only test on identity.
 * The lexer always produces tokens with a dummy ID because it does not
 * know which entity is the current one. The entity layer replaces the dummy
 * ID with the actual ID. The parser checks that the IDs of pairs such as
 * Decl_element and Decl_rangle are the same; otherwise a Validation_error
 * is produced.
 *
 * Revision 1.9  1999/08/15 20:42:01  gerd
 * 	Corrected a misleading message.
 *
 * Revision 1.8  1999/08/15 20:37:34  gerd
 * 	Improved error messages.
 * 	Bugfix: While parsing document entities, the subclass document_entity is
 * now used instead of external_entity. The rules in document entities are a bit
 * stronger.
 *
 * Revision 1.7  1999/08/15 14:03:59  gerd
 * 	Empty documents are not allowed.
 * 	"CDATA section not allowed here" is a WF_error, not a Validation_
 * error.
 *
 * Revision 1.6  1999/08/15 02:24:19  gerd
 * 	Removed some grammar rules that were used for testing.
 * 	Documents without DTD can now have arbitrary elements (formerly
 * they were not allowed to have any element).
 *
 * Revision 1.5  1999/08/14 22:57:20  gerd
 * 	It is allowed that external entities are empty because the
 * empty string is well-parsed for both declarations and contents. Empty
 * entities can be referenced anywhere because the references are replaced
 * by nothing. Because of this, the Begin_entity...End_entity brace is only
 * inserted if the entity is non-empty. (Otherwise references to empty
 * entities would not be allowed anywhere.)
 * 	As a consequence, the grammar has been changed such that a
 * single Eof is equivalent to Begin_entity,End_entity without content.
 *
 * Revision 1.4  1999/08/14 22:20:01  gerd
 *         The "config" slot has now a component "warner" which is
 * an object with a "warn" method. This is used to warn about characters
 * that cannot be represented in the Latin 1 alphabet.
 *         Furthermore, there is a new component "debugging_mode".
 *         Some Parse_error exceptions have been changed into Validation_error.
 *         The interfaces of functions/classes imported from other modules
 * have changed; the invocations have been adapted.
 *         Contents may contain CDATA sections that have been forgotten.
 *
 * Revision 1.3  1999/08/11 15:00:41  gerd
 * 	The Begin_entity ... End_entity brace is also possible in
 * 'contents'.
 * 	The configuration passed to the parsing object contains always
 * the resolver that is actually used.
 *
 * Revision 1.2  1999/08/10 21:35:12  gerd
 * 	The XML/encoding declaration at the beginning of entities is
 * evaluated. In particular, entities have now a method "xml_declaration"
 * which returns the name/value pairs of such a declaration. The "encoding"
 * setting is interpreted by the entity itself; "version", and "standalone"
 * are interpreted by Markup_yacc.parse_document_entity. Other settings
 * are ignored (this does not conform to the standard; the standard prescribes
 * that "version" MUST be given in the declaration of document; "standalone"
 * and "encoding" CAN be declared; no other settings are allowed).
 * 	TODO: The user should be warned if the standard is not exactly
 * fulfilled. -- The "standalone" property is not checked yet.
 *
 * Revision 1.1  1999/08/10 00:35:52  gerd
 * 	Initial revision.
 *
 *
 *)
